BUCKET_SIZE = 10


def bucket_sort(array: list[int]) -> list[int]:
    """
    Time Complexity of Solution:
    Worst Case: occurs when all the elements are placed in a single bucket.
    The overall performance would then be dominated by the algorithm used to sort each
    bucket. In this case, O(n log n), because of TimSort.

    Average Case: O(n + (n^2)/k + k), where k is the number of buckets.
    If k = O(n), time complexity is O(n).

    Runtime: O(n)
    """
    if not array:
        return array
    min_value, max_value = min(array), max(array)
    bucket_count = (max_value - min_value) // BUCKET_SIZE + 1
    buckets: list[list[int]] = [[] for _ in range(int(bucket_count))]

    for item in array:
        buckets[int((item - min_value) // BUCKET_SIZE)].append(item)

    return sorted(
        buckets[i][j] for i in range(len(buckets)) for j in range(len(buckets[i]))
    )



import random
from collections.abc import Sequence
from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def quick_select(items: Sequence[T], index: int) -> T:
    """
    A Python implementation of the quick select algorithm (also called the kth order
    statistic), which is efficient for calculating the value that would appear in the
    index of a list if it would be sorted, even if it is not already sorted.

    Partition the data into smaller and greater lists in relation to the pivot, then
    recurse.

    Average runtime: O(n)
    Runtime: O(n^2)
    This algorithm is often much better than sorting, which is O(n log n).
    """
    if index < 0 or index >= len(items):
        raise ValueError(f"Index {index} is out of range.")

    pivot = random.choice(items)
    equal, smaller, larger = 0, [], []
    for elem in items:
        if elem == pivot:
            equal += 1
        elif elem < pivot:
            smaller.append(elem)
        else:
            larger.append(elem)

    m = len(smaller)
    if index == m:
        return pivot
    if index < m:
        return quick_select(smaller, index)
    # eliminate all smaller elements than pivot
    return quick_select(larger, index - (m + equal))



from cs.structures import Graph, V


def floyd_warshall_shortest_paths(graph: Graph[V]) -> dict[V, dict[V, float]]:
    """
    Calculates the shortest distance between all vertex pairs using dynamic programming.
    distance[u][v] will contain the shortest distance from vertex u to v.

    1. For all edges from v to n, distance[i][j] = weight(edge(i, j)).
    3. The algorithm then performs distance[i][j] = min(distance[i][j], distance[i][k]
        + distance[k][j]) for each possible pair i, j of vertices.
    4. The above is repeated for each vertex k in the graph.
    5. Whenever distance[i][j] is given a new minimum value, next vertex[i][j] is
        updated to the next vertex[i][k].

    Unlike Dijkstra's Algorithm, the Floyd-Warshall Algorithm has no problems handling
    graphs with negative edge costs.

    Runtime: O(|V|^3) Memory: O(|V|^2)
    """
    dist: dict[V, dict[V, float]] = {}
    for u in graph:
        dist[u] = {}
        for v in graph:
            if u == v:
                dist[u][v] = 0
            elif graph.has_edge(u, v):
                dist[u][v] = graph[u][v].weight
            else:
                dist[u][v] = Graph.INFINITY

    # check vertex k against all other vertices (i, j)
    for k in graph:
        for i in graph:
            for j in graph:
                dist[i][j] = min(dist[i][j], dist[i][k] + dist[k][j])
    return dist



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, unique
from typing import TypeVar

from cs.structures.tree.binary_search_tree import BinarySearchTree, BinaryTreeNode
from cs.util import Comparable, dfield

T = TypeVar("T", bound=Comparable)


@unique
class Color(Enum):
    BLACK = "black"
    RED = "red"


@dataclass(order=True, repr=False, slots=True)
class RedBlackTreeNode(BinaryTreeNode[T]):
    left: RedBlackTreeNode[T] | None = None
    right: RedBlackTreeNode[T] | None = None
    parent: RedBlackTreeNode[T] | None = dfield(None)
    color: Color = dfield(Color.BLACK)
    rank: int = dfield(0)  # Defined here because rank is not updated in BSTs.

    @property
    def grandparent(self) -> RedBlackTreeNode[T] | None:
        """Redefined here because @property cannot be inherited."""
        return None if self.parent is None else self.parent.parent

    @property
    def sibling(self) -> RedBlackTreeNode[T] | None:
        """Redefined here because @property cannot be inherited."""
        if self.parent is None:
            return None
        return self.parent.right if self.parent.left is self else self.parent.left


@dataclass
class RedBlackTree(BinarySearchTree[T]):
    """
    A Red-Black tree, which is a self-balancing BST (binary search tree). This tree has
    similar performance to AVL trees, but the balancing is less strict, so it will
    perform faster for writing/deleting nodes and slower for reading in the average
    case, though, because they're both balanced binary search trees, both will get the
    same asymptotic performance.

    Unless otherwise specified, all asymptotic runtimes are specified in
    terms of the size of the tree.
    """

    root: RedBlackTreeNode[T] | None = None
    size: int = 0

    @staticmethod
    def color(node: RedBlackTreeNode[T] | None) -> Color:
        """Returns the color of a node, allowing for None leaves."""
        return Color.BLACK if node is None else node.color

    @staticmethod
    def opposite_color(node: RedBlackTreeNode[T]) -> Color:
        """Returns the color of a node, allowing for None leaves."""
        return Color.RED if node.color is Color.BLACK else Color.BLACK

    @staticmethod
    def rotate_left(node: RedBlackTreeNode[T]) -> RedBlackTreeNode[T]:
        """
        Rotate the subtree rooted at this node to the left and
        returns the new root to this subtree.
        Performing one rotation can be done in O(1).
        """
        right = node.right
        if right is None:
            raise RuntimeError("There is no left node to rotate to as the new root.")
        node.right = right.left
        if node.right is not None:
            node.right.parent = node
        parent = node.parent
        node.parent = right
        right.left = node
        if parent is not None:
            if node.is_left():
                parent.left = right
            else:
                parent.right = right
        right.parent = parent
        return right

    @staticmethod
    def rotate_right(node: RedBlackTreeNode[T]) -> RedBlackTreeNode[T]:
        """
        Rotate the subtree rooted at this node to the right and
        returns the new root to this subtree.
        Performing one rotation can be done in O(1).
        """
        left = node.left
        if left is None:
            raise RuntimeError("There is no left node to rotate to as the new root.")
        node.left = left.right
        if node.left is not None:
            node.left.parent = node
        parent = node.parent
        node.parent = left
        left.right = node
        if parent is not None:
            if node.is_right():
                parent.right = left
            else:
                parent.left = left
        left.parent = parent
        return left

    def check_correctness(self) -> bool:
        """
        Check the coloring of the tree, and return True iff the tree
        is colored in a way which matches these five properties:
        1. Each node is either red or black.
        2. The root node is black.
        3. All leaves are black.
        4. If a node is red, then both its children are black.
        5. Every path from any node to all of its descendent NIL nodes
            has the same number of black nodes.
        This function runs in O(n) time, because of properties 4 and 5.
        """

        def _check_coloring(node: RedBlackTreeNode[T] | None) -> bool:
            """
            A helper function to recursively check Property 4 of a Red-Black Tree.
            See check_color_properties for more info.
            """
            return node is None or (
                node.color is Color.BLACK
                or self.color(node.left) is self.color(node.right) is Color.BLACK
                and _check_coloring(node.left)
                and _check_coloring(node.right)
            )

        def _black_height(node: RedBlackTreeNode[T] | None) -> int:
            """
            Returns the number of black nodes from this node to the leaves of the tree,
            or -1 if there isn't one such value (the tree is colored incorrectly).
            """
            if node is None:
                return 1
            left = _black_height(node.left)
            right = _black_height(node.right)
            if left != right:
                return -1
            return left + int(node.color is Color.BLACK)

        return self.root is None or (
            self.root.color is Color.BLACK
            and _check_coloring(self.root)
            and _black_height(self.root) != -1
        )

    def insert(self, data: T) -> None:
        """
        Inserts data into the subtree rooted at the root, and then performs any
        rotations necessary to maintain balance.

        This is guaranteed to run in O(log(n)) time.
        """
        self.size += 1
        parent, curr = None, self.root
        while curr is not None:
            parent = curr
            if data == curr.data:
                curr.count += 1
                return
            curr = curr.left if data < curr.data else curr.right

        node = RedBlackTreeNode(data=data, parent=parent)
        if parent is None:
            self.root = node
        elif data < parent.data:
            parent.left = node
        else:
            parent.right = node
        self.insert_repair(node)

    def insert_repair(self, node: RedBlackTreeNode[T]) -> None:
        """Repair the coloring from inserting into a tree."""
        while node.parent is not None:
            # For simplicity, get our parent, sibling, uncle, and grandparent.
            # These are the nodes marked in this diagram:
            #           G
            #          / \
            #         P   A
            #        / \
            #       N   S
            # Here, N is the node itself.
            parent = node.parent
            sibling = node.sibling
            grandparent = node.grandparent
            uncle = parent.sibling

            # If the parent corresponds to a node with one key in the 2-3-4 tree (that
            # is, the parent is a black node), then via the isometry we add ourselves
            # to that node by coloring ourselves red. At that point, we're done.
            # To see if our parent corresponds to a node with one key in the 2-3-4
            # tree, we need to check that
            #   1. the parent is black (if it's red, we're part of a larger node), and
            #   2. the parent has no red children (if it does, then it's part
            #      of a larger node).
            # To do this, we'll find our sibling node (the node across from us under our
            # parent) and confirm that it's not red.

            # If the parent is part of a node with two keys in the 2-3-4 tree, add
            # ourselves to that node. There are several cases to consider here, and
            # they're all symmetric. A node with two keys has one of these shapes,
            # with all possible insertion points marked with an I:
            #          B              B
            #         / \            / \
            #        R   I          I   R
            #       / \                / \
            #      I   I              I   I
            # The commonality is that we would be in one of two cases:
            #    1. We have a black parent and a red sibling.
            #    2. We have a red parent and a black uncle.
            # These two cases function differently. If we're in case 1, we just color
            # ourselves red:
            #         B             B
            #        / \    -->    / \
            #       N   R         R   R
            if self.color(parent) is Color.BLACK and (
                (sibling is None or self.color(sibling) is Color.BLACK)
                or (sibling is not None and self.color(sibling) is Color.RED)
            ):
                node.color = Color.RED
                return

            # There are two subcases here, which correspond to the relative ordering
            # at which the node to insert appears relative to the two other nodes in
            # the 3-node.
            # The first option is the "zig zag" case:
            #       B                   B                   N                B
            #      / \                 / \                 / \              / \
            #     R   B   --->        N   B    --->       R   B    --->    R   R
            #      \     rotate      /        rotate           \  recolor       \
            #       N   N with R    R        N with B           B                B
            # To see whether we're in this case, we have to see whether the orientation
            # of the parent/child and grandparent/parent relations are reversed.
            # The other option is the "zig-zig" case:
            #
            #      B               R                  B
            #     / \             / \                / \
            #    R   B   --->    N   B      --->    R   R
            #   /       rotate        \    recolor       \
            #  N       R with B        B                  B
            if self.color(parent) is Color.RED and (
                uncle is None or self.color(uncle) is Color.BLACK
            ):
                if node.is_left() == parent.is_left():
                    self.rotate_with_parent(parent)
                    parent.color = Color.BLACK
                    node.color = Color.RED
                else:
                    self.rotate_with_parent(node)
                    self.rotate_with_parent(node)
                assert grandparent is not None
                grandparent.color = Color.RED
                return

            # Otherwise, we are inserting into a 4-node. There are several orientations
            # possible here, but with mirroring excluded there are basically two unique
            # insertion points
            #          B              B
            #        /   \          /   \
            #       R     R        R     R
            #      /                \
            #     I                  I
            # We are splitting a node with four keys into a node with two keys, a node
            # with one key, and then kicking one key higher up. This can be done purely
            # by recoloring the nodes and continuing the search from a starred node that
            # is colored black beforehand:
            #          B              B
            #        /   \          /   \
            #       R     R        R     R
            #      /                \
            #     I                  I
            #         vvv            vvvv
            #
            #          #              *
            #        /   \          /   \
            #       B     B        B     B
            #      /                \
            #     R                  R
            # In other words, we just flip the colors of the nodes and propagate the
            # search upward from the grandparent.
            assert uncle is not None
            parent.color = Color.BLACK
            uncle.color = Color.BLACK
            node.color = Color.RED
            node = grandparent  # type: ignore[assignment]

    def rotate_with_parent(self, node: RedBlackTreeNode[T]) -> None:
        """
        Rotate the subtree rooted at this node to the right and
        returns the new root to this subtree.
        Performing one rotation can be done in O(1).
        """
        if node.parent is None:
            raise RuntimeError("Node is the root and has no parent.")

        # Step 1: Do the logic to "locally" rotate the nodes. This repositions the
        # node, its parent, and the middle child. However, it leaves the parent
        # pointers of these nodes unmodified; we'll handle that later.
        child = None
        if node.is_left():
            # Rotate right
            child = node.right
            node.right = node.parent
            node.parent.left = child
            node.parent.rank = node.parent.rank - node.rank - 1
        else:
            # Rotate left
            child = node.left
            node.left = node.parent
            node.parent.right = child
            node.rank = node.parent.rank + node.rank + 1

        # Step 2: Make the node's grandparent now point at it.
        grandparent = node.grandparent
        if grandparent is None:
            self.root = node
        else:
            if node.parent.is_left():
                grandparent.left = node
            else:
                grandparent.right = node

        # Step 3: Update parent pointers.
        #  1. The child node that got swapped needs its parent updated.
        #  2. The node we rotated now has a new parent.
        #  3. The node's old parent now points to the node we rotated.
        # We have to be super careful about this, though, because some of these
        # nodes might not exist and we need to not lose any pointers.
        if child is not None:
            child.parent = node.parent
        old_parent = node.parent
        node.parent = old_parent.parent
        old_parent.parent = node

    def select(self, rank: int) -> T:
        """Takes in an integer rank and returns the rank-th order statistic."""
        curr = self.root
        while curr is not None:
            if rank == curr.rank:
                return curr.data
            if rank < curr.rank:
                curr = curr.left
            else:
                rank -= curr.rank + 1
                curr = curr.right
        raise RuntimeError("Not enough elements in tree.")

    def rank_of(self, data: T) -> int:
        """
        Inverse of select; takes in a node in the tree and returns the number
        of elements in the tree smaller than that key.
        """
        curr = self.root
        total = 0
        while curr is not None:
            if data == curr.data:
                return total + curr.rank
            if data < curr.data:
                curr = curr.left
            else:
                total += curr.rank + 1
                curr = curr.right
        return total

    # TODO: fix remove()
    # def remove(self, data: T) -> None:
    #     def _remove(node: RedBlackTreeNode[T]) -> RedBlackTreeNode[T] | None:
    #         nonlocal data
    #         if node.data == data:
    #             if node.left is not None and node.right is not None:
    #                 # It's easier to balance a node with at most one child,
    #                 # so we replace this node with the greatest one less than
    #                 # it and remove that.
    #                 value = node.left.get_max()
    #                 node.data = value
    #                 data = value
    #                 _remove(node.left)
    #             else:
    #                 # This node has at most one non-None child, so we don't
    #                 # need to replace
    #                 child = node.left or node.right
    #                 if node.color is Color.RED:
    #                     # This node is red, and its child is black
    #                     # The only way this happens to a node with one child
    #                     # is if both children are None leaves.
    #                     # We can just remove this node and call it a day.
    #                     if node.is_left():
    #                         node.parent.left = None
    #                     else:
    #                         node.parent.right = None
    #                 else:
    #                     # The node is black
    #                     if child is None:
    #                         # This node and its child are black
    #                         if node.parent is None:
    #                             # The tree is now empty
    #                             return

    #                         self._remove_repair(node)
    #                         if node.is_left:
    #                             node.parent.left = None
    #                         else:
    #                             node.parent.right = None
    #                         node.parent = None
    #                     else:
    #                         # This node is black and its child is red
    #                         # Move the child node here and make it black
    #                         node.data = child.data
    #                         node.left = child.left
    #                         node.right = child.right
    #                         if node.left:
    #                             node.left.parent = self.root
    #                         if node.right:
    #                             node.right.parent = self.root
    #         elif node.data > data:
    #             if node.left:
    #                 _remove(node.left)
    #         else:
    #             if node.right:
    #                 _remove(node.right)

    #     _remove(self.root)

    # def _remove_repair(self, node: RedBlackTreeNode[T]) -> None:
    #         #  if self.color(node.sibling) is Color.RED:
    #         node.sibling.color = Color.BLACK
    #         node.parent.color = Color.RED
    #         if node.is_left:
    #             self.rotate_left(node.parent)
    #         else:
    #             self.rotate_right(node.parent)
    #     if (
    #         self.color(node.parent) is Color.BLACK
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.left) is Color.BLACK
    #         and self.color(node.sibling.right) is Color.BLACK
    #     ):
    #         node.sibling.color = Color.RED
    #         self._remove_repair(node.parent)
    #         return
    #     if (
    #         self.color(node.parent) is Color.RED
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.left) is Color.BLACK
    #         and self.color(node.sibling.right) is Color.BLACK
    #     ):
    #         node.sibling.color = Color.RED
    #         node.parent.color = Color.BLACK
    #         return
    #     if (
    #         node.is_left
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.right) is Color.BLACK
    #         and self.color(node.sibling.left) is Color.RED
    #     ):
    #         self.rotate_right(node.sibling)
    #         node.sibling.color = Color.BLACK
    #         node.sibling.right.color = Color.RED
    #     if (
    #         node.is_right
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.right) is Color.RED
    #         and self.color(node.sibling.left) is Color.BLACK
    #     ):
    #         self.rotate_left(node.sibling)
    #         node.sibling.color = Color.BLACK
    #         node.sibling.left.color = Color.RED
    #     if (
    #         node.is_left
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.right) is Color.RED
    #     ):
    #         self.rotate_left(node.parent)
    #         node.grandparent.color = node.parent.color
    #         node.parent.color = Color.BLACK
    #         node.parent.sibling.color = Color.BLACK
    #     if (
    #         node.is_right
    #         and self.color(node.sibling) is Color.BLACK
    #         and self.color(node.sibling.left) is Color.RED
    #     ):
    #         self.rotate_right(node.parent)
    #         node.grandparent.color = node.parent.color
    #         node.parent.color = Color.BLACK
    #         node.parent.sibling.color = Color.BLACK



from cs.algorithms import bellman_ford_shortest_paths, dijkstra_shortest_paths
from cs.structures import Edge, Graph, Node, V


def johnsons_shortest_paths(graph: Graph[V]) -> dict[V, dict[V, float]]:
    """
    An implementation of Johnson's all-pairs shortest paths algorithm.  This algorithm
    is remarkable in that it combines two single-source shortest path algrithms -
    Dijkstra's algorithm and the Bellman-Ford algorithm - into a single algorithm for
    all-pairs shortest paths that ends up being more efficient than the Floyd-Warshall
    all-pairs shortest paths algorithm when run on sparse graphs.

    The idea behind Johnson's algorithm is to enforce that all the edge costs in the
    graph are nonnegative by computing a new potential function h(v) for each node in
    the graph, and then creating a new edge cost for each edge. With this new graph, we
    can use Dijkstra's to get the shortest paths.

    Runtime: O(|V||E| + |V|^2 log |V|)
    """
    # Create augmented graph G' that will be fed into the Bellman-Ford algorithm by
    # copying the graph and adding an extra node. Adding a new directed source node
    # ensures that there will be no negative cycles in this new graph. We convert the
    # graph into a Node[tuple[V]] type in order to ensure we can add a new, unique node.
    aug_graph = Graph.from_graph(graph, node_fn=lambda x: Node((x,)))
    source_node = Node(())
    aug_graph.add_node(source_node)
    for u in aug_graph:
        if u != source_node:
            aug_graph.add_edge(source_node, u, 0)

    # Convert the Bellman-Ford output of Node[tuple[V]] back into the original type.
    bellman_costs = {
        node.data[0]: cost
        for node, cost in bellman_ford_shortest_paths(aug_graph, source_node).items()
        if node.data
    }

    # Given a graph and a potential function on that graph (maps nodes to their
    # potentials), produces a new graph whose edges are weighted by the potential.
    reweighted_graph = Graph.from_graph(
        graph,
        edge_fn=lambda e: Edge(
            e.start, e.end, e.weight + bellman_costs[e.start] - bellman_costs[e.end]
        ),
    )

    # Run Dijkstra's algorithm over every node in the reweighted graph
    # to get the transformed shortest path costs.
    return {
        u: {
            v: cost + bellman_costs[v] - bellman_costs[u]
            for v, cost in dijkstra_shortest_paths(reweighted_graph, u).items()
        }
        for u in graph
    }



from dataclasses import dataclass, field
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(slots=True)
class Queue(Generic[T]):
    """You should probably use the Python built-in list or deque instead."""

    _queue: list[T] = field(default_factory=list)

    def __bool__(self) -> bool:
        return bool(self._queue)

    def __len__(self) -> int:
        return len(self._queue)

    def __contains__(self, item: T) -> bool:
        return item in self._queue

    def enqueue(self, item: T) -> None:
        self._queue.append(item)

    def dequeue(self) -> T:
        if not self._queue:
            raise IndexError
        return self._queue.pop(0)

    def peek(self) -> T:
        if not self._queue:
            raise IndexError
        return self._queue[0]



from __future__ import annotations

from collections.abc import Sequence
from typing import TypeVar

from cs.structures import BinarySearchTree, BinaryTreeNode, Graph
from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def build_optimal_bst(
    nodes: Sequence[BinaryTreeNode[T]],
) -> tuple[BinarySearchTree[T], float]:
    """
    This function implements an optimal binary search tree-building dynamic programming
    algorithm that delivers O(n^2) performance.

    The goal of the optimal BST problem is to build a low-cost BST for a given set of
    nodes, each with its own key and frequency. The frequency of the node is defined as
    how many times the node is being searched. The search cost of binary search tree is
    given by the formula:

    cost(1, n) = sum_{i=1}^n [node_i_freq * (depth(node_i) + 1)]

    where n is number of nodes in the BST. The characteristic of low-cost BSTs is having
    a faster overall search time than other implementations. The reason for their fast
    search time is that the nodes with high frequencies will be placed near the root of
    the tree, while nodes with low frequencies will be placed near the leaves of the
    tree, thus reducing search time in the most frequent instances.
    """

    def _build_tree(
        cost_table: list[list[int]],
        orig_nodes: list[BinaryTreeNode[T]],
        i: int,
        j: int,
        parent: BinaryTreeNode[T] | None = None,
    ) -> BinaryTreeNode[T] | None:
        if i > j or i < 0 or j > len(cost_table) - 1:
            return None

        node = cost_table[i][j]
        tree_node = orig_nodes[node]
        tree_node.parent = parent
        tree_node.left = _build_tree(cost_table, orig_nodes, i, node - 1, tree_node)
        tree_node.right = _build_tree(cost_table, orig_nodes, node + 1, j, tree_node)
        return tree_node

    sorted_nodes = sorted(nodes, key=lambda node: node.data)
    n = len(sorted_nodes)
    freqs = [sorted_nodes[i].hits for i in range(n)]

    # dp stores the overall minimized tree cost. For each key, cost = frequency of key.
    # sums stores the sum of frequencies of nodes between i and j, inclusive.
    # root stores tree roots that will be used for constructing binary search tree.
    dp = [[freqs[i] if i == j else 0 for j in range(n)] for i in range(n)]
    sums = [list(d) for d in dp]
    root = [[i if i == j else 0 for j in range(n)] for i in range(n)]

    for interval_length in range(2, n + 1):
        for i in range(n - interval_length + 1):
            j = i + interval_length - 1
            dp[i][j] = Graph.INT_INFINITY
            sums[i][j] = sums[i][j - 1] + freqs[j]

            # Apply Knuth's optimization (without optimizing: for r in range(i, j + 1))
            for r in range(root[i][j - 1], root[i + 1][j] + 1):  # r is a root
                optimal_left_cost = dp[i][r - 1] if r != i else 0
                optimal_right_cost = dp[r + 1][j] if r != j else 0
                cost = optimal_left_cost + sums[i][j] + optimal_right_cost
                if dp[i][j] > cost:
                    dp[i][j] = cost
                    root[i][j] = r

    tree = BinarySearchTree[T]()
    tree.root = _build_tree(root, sorted_nodes, 0, n - 1)
    tree.size = n
    return tree, dp[0][n - 1]



import math


def fibonacci_recursive(n: int) -> list[int]:
    cache: dict[int, int] = {0: 0, 1: 1}

    def _fib(n: int) -> int:
        if n in cache:
            return cache[n]
        result = _fib(n - 1) + _fib(n - 2)
        cache[n] = result
        return result

    _ = _fib(n - 1)
    return list(cache.values())


def fibonacci_iterative(n: int) -> list[int]:
    seq_out = [0, 1]
    a, b = 0, 1
    for _ in range(2, n):
        a, b = b, a + b
        seq_out.append(b)
    return seq_out


def fibonacci_formula(n: int) -> list[int]:
    """
    Uses Binet's formula to calculate the fibonacci sequence.
    This formula does not yield correct results past n = 72 because of rounding errors.
    """
    seq_out = [0, 1]
    sqrt_5 = math.sqrt(5)
    phi_1 = (1 + sqrt_5) / 2
    phi_2 = (1 - sqrt_5) / 2
    for i in range(2, n):
        temp_out = (phi_1**i - phi_2**i) / sqrt_5
        seq_out.append(int(temp_out))
    return seq_out



from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def insertion_sort(array: list[T]) -> list[T]:
    """
    Insertion sort algorithm implementation.

    Runtime: O(n^2)
    """
    for i in range(1, len(array)):
        j = i
        while j > 0 and array[j - 1] > array[j]:
            array[j], array[j - 1] = array[j - 1], array[j]
            j -= 1

    return array



from __future__ import annotations

from cs.structures import FibonacciHeap, Graph, V


def prims_mst(graph: Graph[V], start_node: V | None = None) -> Graph[V]:
    """
    Given a connected, undirected graph with real-valued edge costs, returns an MST of
    that graph.

    The key difference between Dijkstra's algorithm and Prim's algorithm is that
    Dijkstra's algorithm creates a shortest-path tree from the source node, while Prim's
    algorithm builds an MST from the source node.

    First, we do O(|V|) insertions into the heap. We then do O(|V|) dequeues (since we
    only want a total of |V| - 1 edges). These dequeues take a total of O(|V| log |V)
    time, though any one dequeue might take much more than that. Finally, on each
    dequeue, we scan all outgoing edges from the dequeued node. Since we never consider
    a node twice, the total number of edges visited must be twice the number of edges in
    the graph, since each edge will be visited once from each endpoint, which is O(|E|)
    addiional time.

    Runtime: O(|E| + |V| log |V|)
    """
    mst = Graph[V](is_directed=False)
    if not graph:
        return mst

    heap = FibonacciHeap[V]()
    start = start_node or next(iter(graph))
    mst.add_node(start)
    _add_outgoing_edges(graph, start, mst, heap)

    for _ in range(len(graph) - 1):
        # The algorithm guarantees that we now have the shortest distance to u.
        u, _ = heap.dequeue()
        v, weight = _min_cost_endpoint(u, graph, mst)
        mst.add_node(u)
        mst.add_edge(u, v, weight)
        _add_outgoing_edges(graph, u, mst, heap)
    return mst


def _add_outgoing_edges(
    graph: Graph[V], u: V, mst: Graph[V], heap: FibonacciHeap[V]
) -> None:
    """
    Given a node, updates the priorities of adjacent nodes by following its edges.
    """
    for v, e in graph[u].items():
        if v in mst:
            continue
        if v not in heap:
            heap.enqueue(v, e.weight)
        elif heap[v].priority > e.weight:
            heap.decrease_key(v, e.weight)


def _min_cost_endpoint(node: V, graph: Graph[V], mst: Graph[V]) -> tuple[V, float]:
    """
    Given a node in the source graph and a set of nodes that we've visited
    so far, returns the minimum-cost edge from that node to some node that
    has been visited before.
    """
    end = None
    least_cost = Graph.INFINITY
    for neighbor, edge in graph[node].items():
        if neighbor in mst and edge.weight < least_cost:
            end = neighbor
            least_cost = edge.weight
    if end is None:
        raise AssertionError("Since we dequeued this node, it has at least 1 neighbor.")
    return end, least_cost



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import Any, ClassVar, Generic, NamedTuple, ParamSpec, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")
P = ParamSpec("P")
R = TypeVar("R")


class CacheInfo(NamedTuple):
    hits: int
    misses: int
    capacity: int
    current_size: int


class LRUCache(Generic[KT, VT]):
    """Uses a regular dictionary and the fact that dictionaries are ordered."""

    # Class variable maps the decorator functions to their respective instance
    # ClassVar[dict[Callable[..., Any], LRUCache[KT, VT]]]
    decorator_instance_map: ClassVar[dict[Callable[..., Any], LRUCache[Any, Any]]] = {}

    def __init__(self, max_capacity: int) -> None:
        self.cache: dict[KT, VT] = {}
        self.max_capacity = max_capacity
        self.hits = 0
        self.misses = 0

    def __getitem__(self, key: KT) -> VT | None:
        if key in self.cache:
            self.hits += 1
            # Equivalent to OrderedDict:  self.cache.move_to_end(key)
            self.cache[key] = self.cache.pop(key)
            return self.cache[key]

        self.misses += 1
        return None

    def __setitem__(self, key: KT, val: VT) -> None:
        if key in self.cache:
            del self.cache[key]
        self.cache[key] = val
        if len(self.cache) > self.max_capacity:
            # Equivalent to OrderedDict:  self.cache.popitem(last=False)
            self.cache.pop(next(iter(self.cache)))

    def __bool__(self) -> bool:
        return bool(self.cache)

    def __len__(self) -> int:
        return len(self.cache)

    def __contains__(self, key: KT) -> bool:
        return key in self.cache

    def __repr__(self) -> str:
        return str(
            CacheInfo(self.hits, self.misses, self.max_capacity, len(self.cache))
        )

    @classmethod
    def lru_cache(cls, size: int = 128) -> Callable[[Callable[P, R]], Callable[P, R]]:
        def cache_decorator_inner(func: Callable[P, R]) -> Callable[P, R]:
            @wraps(func)
            def cache_decorator_wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
                if func not in cls.decorator_instance_map:
                    cls.decorator_instance_map[func] = LRUCache(size)

                result = cls.decorator_instance_map[func][args]
                if result is None:
                    result = func(*args, **kwargs)
                    cls.decorator_instance_map[func][args] = result
                return result

            def cache_info() -> LRUCache[KT, VT]:
                return cls.decorator_instance_map[func]

            cache_decorator_wrapper.cache_info = cache_info  # type: ignore[attr-defined]  # noqa: E501
            return cache_decorator_wrapper

        return cache_decorator_inner


def lru_cache(size: int = 128) -> Callable[[Callable[P, R]], Callable[P, R]]:
    return LRUCache.lru_cache(size)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from cs.util import Comparable, dfield, formatter

T = TypeVar("T", bound=Comparable)


@dataclass(order=True, repr=False, slots=True)
class TreeNode(Generic[T]):
    """
    For inheritance to type-check properly, we will need to re-define any TreeNode
    member variables in subclasses.

    We define all augmented information for nodes within this class. If this information
    is not needed for efficiency reasons, it should be fairly simple to remove any lines
    of code using them. However, the overhead they add is so small that it probably
    wouldn't be worth it.

    count: # of times this node was added
    hits: # of times this node was searched
    rank (RBTree-only): # of nodes smaller than this one
    """

    data: T
    count: int = dfield(1)
    hits: int = dfield(0)

    def __repr__(self) -> str:
        return str(formatter.pformat(self))


@dataclass(init=False)
class Tree(Generic[T]):
    """
    Trees can be extended to have any number of child nodes.

    We separate the BinarySearchTree with the TreeNode class to allow the root
    of the tree to be None, which allows this implementation to type-check.

    We set init=False because a Tree cannot receive any parameters on creation, and we
    omit default values to root and size to prevent Tree from being instantiated.

    Trees cannot use __slots__ because root needs to be assigned on the first insert.
    """

    root: TreeNode[T] | None
    size: int

    def __bool__(self) -> bool:
        return self.root is not None

    def __contains__(self, data: T) -> bool:
        return self.search(data) is not None

    def __iter__(self) -> Iterator[TreeNode[T]]:
        raise NotImplementedError

    def __len__(self) -> int:
        return self.size

    @staticmethod
    def depth(tree: Any) -> int:
        raise NotImplementedError

    def height(self) -> int:
        return self.depth(self.root)

    def clear(self) -> None:
        self.root = None

    def search(self, data: T) -> TreeNode[T] | None:
        raise NotImplementedError

    def insert(self, data: T) -> None:
        raise NotImplementedError

    def remove(self, data: T) -> None:
        raise NotImplementedError

    def is_balanced(self) -> bool:
        raise NotImplementedError

    def max_element(self) -> T:
        raise NotImplementedError

    def min_element(self) -> T:
        raise NotImplementedError

    def traverse(self, method: str = "inorder") -> Iterator[T]:
        raise NotImplementedError

    def select(self, rank: int) -> T:
        raise NotImplementedError

    def rank_of(self, data: T) -> int:
        raise NotImplementedError



from cs.structures import Graph, V


def bellman_ford_shortest_paths(
    graph: Graph[V], start: V, *, check_negative_cycles: bool = True
) -> dict[V, float]:
    """
    Bellman-Ford algorithm for the single-source shortest paths problem.

    The Bellman-Ford algorithm computes the shortest paths from a source node to all
    other nodes in the graph, like Dijkstra's algorithm, however with the added
    guarantee that the Bellman-Ford algorithm works correctly in graphs containing
    negative-cost edges, so long as the graph does not contain a negative cycle (in
    which case the cost of a shortest path may not be well-defined).

    If |E| = O(|V|^2), then running the Floyd-Warshall algorithm makes more sense, to
    get all shortest path pairs rather than just one. However, on sparse graphs where
    |E| = Theta(|V|)) the runtime is O(|V|^2), which is much faster.

    Runtime: O(|V||E|), or worst case, O(|V|^3).
    """
    distances = {v: Graph.INFINITY for v in graph}
    distances[start] = 0

    for _ in range(1, len(graph)):
        temp = distances
        for u in graph:
            for v, e in graph[u].items():
                temp[v] = min(temp[v], distances[u] + e.weight)

        distances, temp = temp, distances

    if check_negative_cycles:
        for edge in graph.edges:
            if distances[edge.end] > distances[edge.start] + edge.weight:
                raise AssertionError("Negative weight cycle exists in the graph.")

    return distances



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto, unique
from typing import Any

from cs.structures.tree.binary_search_tree import BinaryTreeNode

MAX_HEIGHT = 1000
INFINITY = 1 << 8  # ~1 million


@unique
class ParentDirection(Enum):
    LEFT, ROOT, RIGHT = auto(), auto(), auto()


@dataclass
class AsciiNode:
    left: AsciiNode | None = None
    right: AsciiNode | None = None
    edge_length: int = 0  # length of the edge from this node to its children
    height: int = 0
    parent_dir: ParentDirection = ParentDirection.ROOT
    label: str = ""  # max supported unit32 in dec, 10 digits max


def build_ascii_tree_recursive(
    t: BinaryTreeNode[Any] | None, direction: ParentDirection = ParentDirection.ROOT
) -> AsciiNode | None:
    if t is None:
        return None

    left_side = build_ascii_tree_recursive(t.left, ParentDirection.LEFT)
    right_side = build_ascii_tree_recursive(t.right, ParentDirection.RIGHT)
    return AsciiNode(
        left=left_side, right=right_side, parent_dir=direction, label=str(t.data)
    )


def compute_profile(
    node: AsciiNode | None,
    lprofile: list[int],
    rprofile: list[int],
    xy: tuple[int, int] = (0, 0),
    left_side: bool = True,
) -> None:
    """
    Fills in the lprofile array for the given tree. It assumes that the center of
    the label of the root of this tree is located at a position (x,y) and assumes
    that the edge_length fields have been computed for this tree.
    """
    if node is None:
        return

    x, y = xy
    if left_side:
        isleft = int(node.parent_dir is ParentDirection.LEFT)
        lprofile[y] = min(lprofile[y], x - ((len(node.label) - isleft) // 2))
    else:
        notleft = int(node.parent_dir is not ParentDirection.LEFT)
        rprofile[y] = max(rprofile[y], x + ((len(node.label) - notleft) // 2))

    if (node.left if left_side else node.right) is not None:
        for i in range(1, min(node.edge_length + 1, MAX_HEIGHT - y)):
            if left_side:
                lprofile[y + i] = min(lprofile[y + i], x - i)
            else:
                rprofile[y + i] = max(rprofile[y + i], x + i)

    compute_profile(
        node.left,
        lprofile,
        rprofile,
        (x - node.edge_length - 1, y + node.edge_length + 1),
        left_side,
    )
    compute_profile(
        node.right,
        lprofile,
        rprofile,
        (x + node.edge_length + 1, y + node.edge_length + 1),
        left_side,
    )


def compute_edge_lengths(
    node: AsciiNode | None, lprofile: list[int], rprofile: list[int]
) -> None:
    """
    This function fills in the edge_length and height fields of the specified tree.
    """
    if node is None:
        return
    compute_edge_lengths(node.left, lprofile, rprofile)
    compute_edge_lengths(node.right, lprofile, rprofile)

    # first fill in the edge_length of node
    if node.left is None and node.right is None:
        node.edge_length = 0
    else:
        hmin = 0
        if node.left is not None:
            for i in range(min(node.left.height, MAX_HEIGHT)):
                rprofile[i] = -INFINITY
            compute_profile(node.left, lprofile, rprofile, left_side=False)
            hmin = node.left.height

        if node.right is not None:
            for i in range(min(node.right.height, MAX_HEIGHT)):
                lprofile[i] = INFINITY
            compute_profile(node.right, lprofile, rprofile, left_side=True)
            # If left is None then this value is 0.
            hmin = min(node.right.height, hmin)

        delta = max([4] + [4 + rprofile[i] - lprofile[i] for i in range(hmin)])
        # If the node has two children of height 1, then we allow the
        # two leaves to be within 1, instead of 2
        if (
            (node.left is not None and node.left.height == 1)
            or (node.right is not None and node.right.height == 1)
        ) and delta > 4:
            delta -= 1
        node.edge_length = ((delta + 1) // 2) - 1

    # now fill in the height of node
    h = 1
    if node.left is not None:
        h = max(node.left.height + node.edge_length + 1, h)
    if node.right is not None:
        h = max(node.right.height + node.edge_length + 1, h)
    node.height = h


def print_level(
    node: AsciiNode | None, x: int, level: int, result: list[str], print_next: int = 0
) -> int:
    """
    This function prints the given level of the given tree, assuming that the node
    has the given x coordinate. print_next is the x coordinate of the next char
    printed, used for printing the next node in the same level.
    """
    if node is None:
        return print_next

    if level == 0:
        isleft = int(node.parent_dir == ParentDirection.LEFT)
        spaces = x - print_next - ((len(node.label) - isleft) // 2)
        print_next += spaces + len(node.label)
        result.append(f"{' ' * spaces}{node.label}")

    elif node.edge_length >= level:
        if node.left is not None:
            spaces = x - print_next - level
            print_next += spaces + 1
            result.append(f"{' ' * spaces}/")
        if node.right is not None:
            spaces = x - print_next + level
            print_next += spaces + 1
            result.append(f"{' ' * spaces}\\")

    else:
        print_next = print_level(
            node.left,
            x - node.edge_length - 1,
            level - node.edge_length - 1,
            result,
            print_next,
        )
        print_next = print_level(
            node.right,
            x + node.edge_length + 1,
            level - node.edge_length - 1,
            result,
            print_next,
        )
    return print_next


def draw_tree(t: BinaryTreeNode[Any] | None) -> str:
    if t is None:
        return ""
    proot = build_ascii_tree_recursive(t)
    assert proot is not None

    lprofile = [0] * MAX_HEIGHT
    rprofile = [0] * MAX_HEIGHT

    compute_edge_lengths(proot, lprofile, rprofile)
    if proot.height >= MAX_HEIGHT:
        raise RuntimeWarning(
            f"This tree is taller than {MAX_HEIGHT}, and may be drawn incorrectly."
        )

    for i in range(min(proot.height, MAX_HEIGHT)):
        lprofile[i] = INFINITY

    compute_profile(proot, lprofile, rprofile, left_side=True)
    xmin = min([0] + [lprofile[i] for i in range(min(proot.height, MAX_HEIGHT))])

    result: list[str] = []
    for i in range(proot.height):
        print_level(proot, -xmin, i, result)
        result.append("\n")

    return "".join(result)



import math

from .precomputed_rmq import PrecomputedRMQ
from .rmq import RMQ
from .sparse_table_rmq import SparseTableRMQ


class FischerHeunRMQ(RMQ):
    def __init__(self, elems: list[int]) -> None:
        super().__init__(elems)
        self.block_size = max(1, math.floor(math.log2(len(elems))) // 4)
        self.block_mins = []
        block_min_vals = []
        for i in range(0, len(elems), self.block_size):
            curr_min = i
            for j in range(i + 1, min(i + self.block_size, len(elems))):
                curr_min = self.return_smaller_index(j, curr_min)
            self.block_mins.append(curr_min)
            block_min_vals.append(elems[curr_min])

        self.summary_rmq = SparseTableRMQ(block_min_vals)
        self.block_index_to_cart = []  # [None] * (4 ** self.block_size)
        self.cart_to_rmq: dict[int, RMQ] = {}
        for i in range(math.ceil(len(elems) / self.block_size)):
            start = i * self.block_size
            current_range = self.elems[start : min(len(elems), start + self.block_size)]
            cartesian_num = self.calc_cart_num(current_range)
            self.block_index_to_cart.append(cartesian_num)
            if cartesian_num not in self.cart_to_rmq:
                self.cart_to_rmq[cartesian_num] = PrecomputedRMQ(current_range)

    @staticmethod
    def calc_cart_num(arr: list[int]) -> int:
        """Calculate cartesian number."""
        result = ""
        stack: list[int] = []
        for elem in arr:
            while stack and stack[-1] > elem:
                stack.pop()
                result += "0"
            stack.append(elem)
            result += "1"
        return int(result[::-1], 2)

    def rmq(self, low: int, high: int) -> int:
        if low >= high:
            raise RuntimeError("In the range, low must be lower than high.")

        min_index = -1
        if high - low < self.block_size:
            for i in range(low, high):
                min_index = self.return_smaller_index(i, min_index)
            return min_index

        start_block = math.ceil(low / self.block_size)
        end_block = math.floor(high / self.block_size)

        if start_block < end_block:
            block_min_index = self.summary_rmq.rmq(start_block, end_block)
            min_index = self.block_mins[block_min_index]

        if low < start_block * self.block_size:
            block_rmq = self.cart_to_rmq[self.block_index_to_cart[start_block - 1]]
            adjust_factor = (start_block - 1) * self.block_size
            new_index = (
                block_rmq.rmq(low - adjust_factor, len(block_rmq.elems)) + adjust_factor
            )
            min_index = self.return_smaller_index(new_index, min_index)

        if end_block * self.block_size < high:
            block_rmq = self.cart_to_rmq[self.block_index_to_cart[end_block]]
            adjust_factor = end_block * self.block_size
            new_index = block_rmq.rmq(0, high - adjust_factor) + adjust_factor
            min_index = self.return_smaller_index(new_index, min_index)

        return min_index



from __future__ import annotations

import random
from collections.abc import Callable
from dataclasses import dataclass
from typing import ClassVar, Generic, TypeVar, cast

KT = TypeVar("KT")
VT = TypeVar("VT")


@dataclass(slots=True)
class TableEntry(Generic[KT, VT]):
    key: KT
    value: VT

    def __repr__(self) -> str:
        return str(self.key)


@dataclass(slots=True)
class HashTable(Generic[KT, VT]):
    """This implementation assumes there are no duplicate keys."""

    num_buckets: int
    load_factor: float
    capacity: int = 0
    num_elems: int = 0
    _hash_ids: ClassVar[dict[int, int]] = {}

    def __post_init__(self) -> None:
        self.capacity = self.num_buckets

    def __bool__(self) -> bool:
        return self.num_elems > 0

    def __len__(self) -> int:
        return self.num_elems

    def __contains__(self, key: KT) -> bool:
        self.validate_key(key)
        return self._find_key(key) is not None

    def __getitem__(self, key: KT) -> VT:
        if (result := self._find_key(key)) is None:
            raise KeyError
        return result.value

    def __setitem__(self, key: KT, value: VT) -> None:
        self.insert(key, value)

    def __delitem__(self, key: KT) -> None:
        self.remove(key)

    @staticmethod
    def generate_hash_function(n: int | None = None) -> Callable[[KT], int]:
        if n is None:
            n = max(HashTable._hash_ids) + 1 if HashTable._hash_ids else 0
        mask = HashTable._hash_ids.get(n)
        if mask is None:
            mask = HashTable._hash_ids[n] = random.getrandbits(32)
        return lambda x: hash(x) ^ cast(int, mask)

    @staticmethod
    def validate_key(key: KT) -> None:
        if key is None:
            raise ValueError("None cannot be inserted into a HashTable.")

    def insert(self, key: KT, value: VT) -> None:
        raise NotImplementedError

    def remove(self, key: KT) -> None:
        raise NotImplementedError

    def _find_key(self, key: KT) -> TableEntry[KT, VT] | None:
        raise NotImplementedError



import random

from cs.structures import Edge, Graph, Node, V


def kargers_min_cut(orig_graph: Graph[V]) -> set[Edge[V]]:
    """
    Partitions a graph using Karger's Algorithm. Works on directed and undirected
    graphs, but involves random choices, so it does not give consistent outputs.
    Args:
        graph: A dictionary containing adacency lists for the graph.
            Nodes must be strings.
    Returns:
        The cutset of the cut found by Karger's Algorithm.
    """
    graph: Graph[Node[tuple[V, ...]]] = Graph.from_graph(
        orig_graph, node_fn=lambda x: Node((x,))
    )

    while len(graph) > 2:
        edge = random.choice(tuple(graph.edges))

        # Contract edge (u, v) to new node uv
        uv = Node(edge.start.data + edge.end.data)

        uv_neighbors = graph[edge.start] | graph[edge.end]
        del uv_neighbors[edge.start]
        del uv_neighbors[edge.end]

        graph.add_node(uv)
        for neighbor in uv_neighbors:
            graph.add_edge(uv, neighbor)
            if graph.is_directed:
                graph.add_edge(neighbor, uv)

        # Remove nodes u and v.
        graph.remove_node(edge.start)
        graph.remove_node(edge.end)

    # Find cutset.
    group1, group2 = graph.nodes
    result_set = set()
    for subnode in group1.data:
        for subneighbor in group2.data:
            if subneighbor in orig_graph[subnode] or subnode in orig_graph[subneighbor]:
                result_set.add(orig_graph[subnode][subneighbor])

    return result_set



from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def quick_sort(array: list[T]) -> list[T]:
    """
    Quick sort algorithm implementation.

    Runtime: O(n log n)
    """
    if len(array) <= 1:
        return array

    pivot = array.pop()
    greater, lesser = [], []
    for element in array:
        if element > pivot:
            greater.append(element)
        else:
            lesser.append(element)
    return quick_sort(lesser) + [pivot] + quick_sort(greater)



from __future__ import annotations

import heapq
from dataclasses import dataclass, field
from pathlib import Path
from typing import cast

from cs.util import dfield


@dataclass(order=True, slots=True)
class HuffmanTreeNode:
    freq: int
    letter: str = field(default="")
    left: HuffmanTreeNode | None = dfield(None)
    right: HuffmanTreeNode | None = dfield(None)
    bitstring: str = field(default="", compare=False)


def parse_file(filepath: Path) -> list[HuffmanTreeNode]:
    """
    Read the file and build a dict of all letters and their
    frequencies, then convert the dict into a list of Letters.
    """
    chars: dict[str, int] = {}
    with open(filepath, encoding="utf-8") as f:
        while c := f.read(1):
            if c in chars:
                chars[c] += 1
            else:
                chars[c] = 1

    queue = [HuffmanTreeNode(freq, ch) for ch, freq in chars.items()]
    heapq.heapify(queue)
    return queue


def build_tree(letters: list[HuffmanTreeNode]) -> HuffmanTreeNode:
    """
    Run through the list of Letters and build the
    min heap for the Huffman Tree.
    """
    while len(letters) > 1:
        left = heapq.heappop(letters)
        right = heapq.heappop(letters)
        node = HuffmanTreeNode(left.freq + right.freq, left=left, right=right)
        heapq.heappush(letters, node)
    return letters[0]


def encode(root: HuffmanTreeNode, bitstring: str) -> list[HuffmanTreeNode]:
    """
    Recursively traverse the Huffman Tree to set each
    letter's bitstring, and return the list of letters.
    """
    if root.left is None or root.right is None:
        root.bitstring = bitstring
        return [root]
    return encode(root.left, bitstring + "0") + encode(root.right, bitstring + "1")


def huffman_compress(
    filepath: Path, output_filepath: Path | None = None
) -> HuffmanTreeNode:
    """
    Parse the file, build the tree, then run through the file
    again, using the list of Letters to find and print out the
    bitstring for each letter.
    """
    # print(f"Huffman Coding of {filepath}: ")
    queue = parse_file(filepath)
    root = build_tree(queue)
    letters = encode(root, "")

    encoding = ""
    with open(filepath, encoding="utf-8") as f:
        while c := f.read(1):
            [byte] = [letter for letter in letters if letter.letter == c]
            encoding += byte.bitstring

    output_path = (
        filepath.with_suffix(".huf") if output_filepath is None else output_filepath
    )
    output_path.write_text(encoding, encoding="utf-8")
    return root


def decode(root: HuffmanTreeNode, filepath: Path) -> str:
    """
    Recursively traverse the Huffman Tree to read each
    letter in the bitstring, and return the decoded string of letters.
    """
    output = ""
    curr = root
    with open(filepath, encoding="utf-8") as f:
        while bit := f.read(1):
            if bit not in ("0", "1"):
                raise ValueError(
                    f"Input bitstring contained character other than 0 or 1: {bit}"
                )
            if curr.letter:
                output += curr.letter
                curr = root
            if bit == "0":
                curr = cast(HuffmanTreeNode, curr.left)
            elif bit == "1":
                curr = cast(HuffmanTreeNode, curr.right)
    output += curr.letter
    return output


def huffman_decompress(filepath: Path, root: HuffmanTreeNode) -> str:
    """
    Parse the file, then use the input to find and print
    out the letter for each bitstring.
    """
    output = decode(root, filepath)
    # print(f"Huffman Decoding of {filepath}: {output}")
    return output



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(repr=False, slots=True)
class LinkedListNode(Generic[T]):
    data: T
    next: LinkedListNode[T] | None = None

    def __repr__(self) -> str:
        return f"({self.data}) -> {self.next}"


@dataclass(init=False)
class LinkedList(Generic[T]):
    head: LinkedListNode[T] | None

    def __init__(self) -> None:
        self.head = None
        self.size = 0

    def __len__(self) -> int:
        return self.size

    def __iter__(self) -> Iterator[LinkedListNode[T]]:
        node = self.head
        while node is not None:
            yield node
            node = node.next

    def __bool__(self) -> bool:
        return self.head is not None

    def __contains__(self, item: T) -> bool:
        curr = self.head
        while curr is not None:
            if curr.data == item:
                return True
            curr = curr.next
        return False

    def __getitem__(self, index: int) -> T:
        current = self.head
        if current is None:
            raise IndexError("List is empty")
        for _ in range(index):
            if current.next is None:
                raise IndexError("Index out of range.")
            current = current.next
        return current.data

    def __setitem__(self, index: int, data: T) -> None:
        current = self.head
        if current is None:
            raise IndexError("List is empty")
        for _ in range(index):
            if current.next is None:
                raise IndexError("Index out of range.")
            current = current.next
        current.data = data

    @classmethod
    def from_list(cls, lst: list[T]) -> LinkedList[T]:
        linked_lst = cls()
        for item in reversed(lst):
            linked_lst.insert(item)
        return linked_lst

    def insert(self, data: T, index: int = 0) -> None:
        """Inserts data to the front of the list, or at the specified index."""
        if index < 0 or index > self.size:
            raise IndexError
        self.size += 1
        if index == 0:
            self.head = LinkedListNode(data, self.head)
            return

        curr = self.head
        prev = None
        for _ in range(index):
            prev = curr
            if curr is None:
                raise ValueError
            curr = curr.next
        if prev is not None:
            prev.next = LinkedListNode(data, curr)

    def remove(self, data: T) -> None:
        if self.head is None:
            raise Exception("List is empty")

        if self.head.data == data:
            self.head = self.head.next
            return

        prev = self.head
        for node in self:
            if node.data == data:
                prev.next = node.next
                return
            prev = node

        raise Exception("Node not found")

    def add_to_end(self, data: T) -> None:
        if self.head is None:
            self.head = LinkedListNode(data)
            return

        curr = self.head
        while curr.next is not None:
            curr = curr.next
        curr.next = LinkedListNode(data)

    def remove_last(self) -> LinkedListNode[T] | None:
        """Deletes the last element of a linked list using only self.head."""

        def _remove_last(lst: LinkedListNode[T] | None) -> LinkedListNode[T] | None:
            if lst is None or lst.next is None:
                return None
            lst.next = _remove_last(lst.next)
            return lst

        return _remove_last(self.head)

    # def swap_nodes(self, node_data_1, node_data_2):
    #     if node_data_1 == node_data_2:
    #         return
    #     else:
    #         node_1 = self.head
    #         while node_1 is not None and node_1.data != node_data_1:
    #             node_1 = node_1.next

    #         node_2 = self.head
    #         while node_2 is not None and node_2.data != node_data_2:
    #             node_2 = node_2.next

    #         if node_1 is None or node_2 is None:
    #             return

    #         node_1.data, node_2.data = node_2.data, node_1.data



import math

from .rmq import RMQ
from .sparse_table_rmq import SparseTableRMQ


class HybridRMQ(RMQ):
    def __init__(self, elems: list[int]) -> None:
        super().__init__(elems)
        self.block_size = max(1, math.floor(math.log2(len(elems))))
        self.block_mins = []
        block_min_vals = []
        for i in range(0, len(elems), self.block_size):
            curr_min = i
            for j in range(i + 1, min(i + self.block_size, len(elems))):
                curr_min = self.return_smaller_index(j, curr_min)
            self.block_mins.append(curr_min)
            block_min_vals.append(elems[curr_min])

        self.summary_rmq = SparseTableRMQ(block_min_vals)

    def rmq(self, low: int, high: int) -> int:
        if low >= high:
            raise RuntimeError("In the range, low must be lower than high.")

        min_index = -1
        if high - low < self.block_size:
            for i in range(low, high):
                min_index = self.return_smaller_index(i, min_index)
            return min_index

        start_block = math.ceil(low / self.block_size)
        end_block = math.floor(high / self.block_size)

        if start_block < end_block:
            block_min_index = self.summary_rmq.rmq(start_block, end_block)
            min_index = self.block_mins[block_min_index]

        for i in range(low, min(high, start_block * self.block_size)):
            min_index = self.return_smaller_index(i, min_index)

        for i in range(max(low, end_block * self.block_size), high):
            min_index = self.return_smaller_index(i, min_index)
        return min_index



from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def bubble_sort(array: list[T]) -> list[T]:
    """
    Bubble sort algorithm implementation.

    Runtime: O(n^2)
    """
    for i in range(len(array) - 1):
        for j in range(len(array) - 1 - i):
            if array[j] > array[j + 1]:
                array[j], array[j + 1] = array[j + 1], array[j]
    return array



from cs.algorithms import breadth_first_search
from cs.structures import Graph, V


def edmonds_karp_max_flow(graph: Graph[V], source: V, sink: V) -> float:
    flow_network = Graph.from_graph(graph)
    max_flow = 0.0

    # Augment the flow while there is path from source to sink
    while path := breadth_first_search(flow_network, source, sink):
        # Find minimum residual capacity of the edges along the path filled by BFS,
        # aka the maximum flow through the path found, by iterating over pairs of nodes.
        path_flow = min(
            flow_network[path[i]][path[i + 1]].weight for i in range(len(path) - 1)
        )
        max_flow += path_flow

        # Update residual capacities of the edges and reverse edges along the path
        for i in range(len(path) - 1):
            u, v = path[i], path[i + 1]
            if u not in flow_network[v]:
                flow_network.add_edge(v, u, 0)
            flow_network[u][v].weight -= path_flow
            flow_network[v][u].weight += path_flow
    return max_flow



def _longest_common_subsequence(s1: str, s2: str) -> int:
    """
    Let m and n be the lengths of two strings.
    Build L[m+1][n+1] from the bottom up.
    Note: L[i][j] contains length of LCS of X[0..i-1] and Y[0..j-1]

    Runtime: O(mn)
    Space Complexity: O(mn)
    """
    m, n = len(s1), len(s2)
    L = [[0] * (n + 1) for i in range(m + 1)]
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if s1[i - 1] == s2[j - 1]:
                L[i][j] = L[i - 1][j - 1] + 1
            else:
                L[i][j] = max(L[i - 1][j], L[i][j - 1])
    return L[m][n]


def longest_common_subsequence(s1: str, s2: str) -> int:
    """
    Space-optimized version of LCS.
    Let m and n be the lengths of two strings.

    Runtime: O(mn)
    Space Complexity: O(min(m, n))
    """
    m, n = len(s1), len(s2)
    if m < n:
        s1, s2 = s2, s1

    L = [0] * (n + 1)
    for a in s1:
        prev_row, prev_row_col = 0, 0
        for j, b in enumerate(s2):
            prev_row, prev_row_col = L[j + 1], prev_row
            if a == b:
                L[j + 1] = prev_row_col + 1
            else:
                L[j + 1] = max(L[j], prev_row)
    return L[-1]



from __future__ import annotations

from dataclasses import dataclass

from .hash_table import KT, VT, HashTable, TableEntry


@dataclass(slots=True)
class LinearProbingEntry(TableEntry[KT, VT]):
    is_dead: bool = False


class LinearProbing(HashTable[KT, VT]):
    def __init__(self, num_buckets: int, load_factor: float = 0.4) -> None:
        super().__init__(num_buckets, load_factor)
        self.table: list[LinearProbingEntry[KT, VT] | None] = [None] * num_buckets

    def __str__(self) -> str:
        result = ""
        for i in range(self.num_buckets):
            entry = self.table[i]
            output = None if entry is None or entry.is_dead else entry.value
            result += f"{i}  |  {output}\n"
        return result

    def insert(self, key: KT, value: VT) -> None:
        self.validate_key(key)
        # Skip check for duplicates because membership check is very slow.

        # Resize table if size exceeds capacity.
        if self.num_elems >= self.load_factor * self.capacity:
            self.capacity *= 2
            self.num_buckets *= 2
            self.num_elems = 0
            table = list(self.table)
            self.table = [None] * self.num_buckets
            for entry in table:
                if entry is not None and not entry.is_dead:
                    self.insert(entry.key, entry.value)

        bucket = hash(key) % self.num_buckets
        while (entry := self.table[bucket]) is not None and not entry.is_dead:
            # Replace existing key
            if entry.key == key:
                entry.value = value
                return
            bucket = (bucket + 1) % self.num_buckets

        self.table[bucket] = LinearProbingEntry(key, value)
        self.num_elems += 1

    def remove(self, key: KT) -> None:
        self.validate_key(key)
        if (entry := self._find_key(key)) is None:
            raise KeyError

        entry.is_dead = True
        self.num_elems -= 1

    def get_elems(self) -> list[tuple[KT, VT]]:
        return [(elem.key, elem.value) for elem in self.table if elem is not None]

    def _find_key(self, key: KT) -> LinearProbingEntry[KT, VT] | None:
        bucket = hash(key) % self.num_buckets
        for i in range(self.num_buckets):
            index = (bucket + i) % self.num_buckets
            entry = self.table[index]
            if entry is None:
                break
            if entry.key == key and not entry.is_dead:
                return entry
        return None



def greatest_common_divisor(a: int, b: int) -> int:
    while b:
        a, b = b, a % b
    return a


def least_common_multiple(a: int, b: int) -> int:
    return a // greatest_common_divisor(a, b) * b



from __future__ import annotations

from cs.structures import Graph, V


def depth_first_search(graph: Graph[V], start: V, end: V) -> list[V]:
    """
    Iterative version of DFS.

    Runtime: O(V + E)
    """
    stack: list[tuple[V, list[V]]] = [(start, [start])]
    visited: set[V] = set()
    while stack:
        vertex, path = stack.pop()
        if vertex == end:
            return path
        if vertex not in visited:
            visited.add(vertex)
            for neighbor in graph[vertex]:
                if neighbor not in visited:
                    stack.append((neighbor, path + [neighbor]))
    return []


def depth_first_search_recursive(graph: Graph[V], start: V, end: V) -> list[V]:
    """
    Recursive version of DFS.

    Runtime: O(V + E)
    """

    def _dfs(curr: V, visited: set[V], path: list[V]) -> list[V]:
        if curr == end:
            return path
        if curr not in visited:
            visited.add(curr)
            for neighbor in graph[curr]:
                if result := _dfs(neighbor, visited, path + [neighbor]):
                    return result
        return []

    return _dfs(start, set(), [start])


def dfs_traversal(graph: Graph[V], start: V, visited: set[V] | None) -> set[V]:
    """
    Explores graph starting with start using a depth-first-search traversal.
    Modifies a visited set in place, and returns a set of connected vertices.

    Runtime: O(V + E)
    """
    if visited is None:
        visited = set()
    visited.add(start)
    connected_nodes = {start}
    for neighbor in graph[start]:
        if neighbor not in visited:
            connected_nodes |= dfs_traversal(graph, neighbor, visited)
    return connected_nodes



from cs.structures import DisjointSet, Graph, V


def kruskals_mst(graph: Graph[V]) -> Graph[V]:
    """
    Kruskal's Algorithm to find an MST.
    Not 100% deterministic because edges with the same weight are arbitrarily ordered.

    Kruskal's algorithm sorts all of the graph's edges in ascending order of size, then
    adds one at a time back to a new graph. It maintains a union-find data structure to
    prevent edge additions that would add a cycle to the new graph. Using a union-find
    structure for this gives a worse runtime worse than the O(|E| + |V| log |V|)
    guarantee of Prim's algorithm. However, Prim's algorithm relies on the often slower
    Fibonacci heap, and so Kruskal's algorithm is often faster in practice.

    Runtime: O(|E| log |V|)
    """
    mst = Graph[V](is_directed=False)
    if len(graph) <= 1:
        return mst

    edge_queue = sorted(graph.edges, reverse=True, key=lambda e: e.weight)
    disjoint_sets = DisjointSet[V]()
    for v in graph:
        mst.add_node(v)
        disjoint_sets.make_set(v)

    num_sets = len(disjoint_sets)
    while num_sets != 1:
        edge = edge_queue.pop()
        orig_num_sets = num_sets
        disjoint_sets.union(edge.start, edge.end)
        num_sets = len(disjoint_sets)
        if num_sets < orig_num_sets:
            mst.add_edge(**edge)
    return mst



from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def merge_sort(array: list[T]) -> list[T]:
    """
    Merge sort algorithm implementation.

    Runtime: O(n log n)
    """

    def merge(left: list[T], right: list[T]) -> list[T]:
        """Merge sort merging function."""
        left_index, right_index = 0, 0
        result = []
        while left_index < len(left) and right_index < len(right):
            if left[left_index] < right[right_index]:
                result.append(left[left_index])
                left_index += 1
            else:
                result.append(right[right_index])
                right_index += 1
        return result + left[left_index:] + right[right_index:]

    if len(array) <= 1:
        return array

    half = len(array) // 2
    left = merge_sort(array[:half])
    right = merge_sort(array[half:])
    return merge(left, right)



def factorial_iterative(n: int) -> int:
    if n < 0:
        raise ValueError("factorial() not defined for negative values")
    value = 1
    for i in range(1, n + 1):
        value *= i
    return value


def factorial_recursive(n: int) -> int:
    if n < 0:
        raise ValueError("factorial() not defined for negative values")
    if n <= 1:
        return 1
    return n * factorial_recursive(n - 1)



def roman_to_int(roman: str) -> int:
    """
    Roman to Integer
    Given a roman numeral, convert it to an integer.
    Input is guaranteed to be within the range from 1 to 3999.
    https://en.wikipedia.org/wiki/Roman_numerals
    """
    vals = {"I": 1, "V": 5, "X": 10, "L": 50, "C": 100, "D": 500, "M": 1000}
    total, place = 0, 0
    while place < len(roman):
        if place + 1 < len(roman) and vals[roman[place]] < vals[roman[place + 1]]:
            total += vals[roman[place + 1]] - vals[roman[place]]
            place += 2
        else:
            total += vals[roman[place]]
            place += 1
    return total



import math
from pprint import pformat

from .rmq import RMQ


class SparseTableRMQ(RMQ):
    def __init__(self, elems: list[int]) -> None:
        super().__init__(elems)
        log_elems = math.floor(math.log2(len(elems))) + 1
        self.rmq_table = [[-1] * len(elems) for _ in range(log_elems)]

        for i in range(len(elems)):
            self.rmq_table[0][i] = i

        for i in range(1, log_elems + 1):
            for j in range(len(elems) - (2**i) + 1):
                index1 = self.rmq_table[i - 1][j]
                index2 = self.rmq_table[i - 1][j + 2 ** (i - 1)]
                self.rmq_table[i][j] = self.return_smaller_index(index1, index2)

    def __repr__(self) -> str:
        # return pformat(np.transpose(np.array(self.rmq_table)))
        return pformat(self.rmq_table)

    def rmq(self, low: int, high: int) -> int:
        if low >= high:
            raise RuntimeError("In the range, low must be lower than high.")

        def get_k(n: int) -> int:
            return 67 - len(bin(-n)) & ~n >> 64

        k = 63 - get_k(high - low)
        index1 = self.rmq_table[k][low]
        index2 = self.rmq_table[k][high - 2**k]
        return self.return_smaller_index(index1, index2)



import dataclasses
import random
from abc import abstractmethod
from pathlib import Path
from typing import Any, Protocol, TypeVar

import prettyprinter  # type: ignore[import]
from prettyprinter.prettyprinter import IMPLICIT_MODULES  # type: ignore[import]

C = TypeVar("C", bound="Comparable")


class Comparable(Protocol):
    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __lt__(self: C, other: C) -> bool:
        pass

    def __gt__(self: C, other: C) -> bool:
        return (not self < other) and self != other

    def __le__(self: C, other: C) -> bool:
        return self < other or self == other

    def __ge__(self: C, other: C) -> bool:
        return not self < other


class Formatter:
    """Wrapper class for PrettyPrinter."""

    def __init__(self, prettyprinter_module: Any) -> None:
        self.pformat = prettyprinter_module.pformat
        self.pprint = prettyprinter_module.pprint


def init_prettyprinter() -> Formatter:
    """Initialize prettyprinter and add all IMPLICIT_MODULES."""
    prettyprinter.install_extras(include={"python"})
    prettyprinter.register_pretty(predicate=dataclasses.is_dataclass)(
        pretty_dataclass_instance
    )
    for filepath in Path("cs").rglob("*.py"):
        module_name = filepath.stem
        if "__" not in module_name:
            prefix = ".".join(filepath.parts[:-1] + (module_name,))
            IMPLICIT_MODULES.add(prefix)
    return Formatter(prettyprinter)


def pretty_dataclass_instance(value: Any, ctx: Any) -> Any:
    cls = type(value)
    kwargs = {}
    for field_def in dataclasses.fields(value):
        # repr is True by default, therefore if this is False, the user
        # has explicitly indicated they don't want to display the field value.
        if not field_def.repr:
            continue

        default = field_def.default
        default_factory: Any = field_def.default_factory
        true_val = getattr(value, field_def.name)
        display_attr = (
            default is default_factory is dataclasses.MISSING
            or (default is not dataclasses.MISSING and default != true_val)
            or (
                default_factory is not dataclasses.MISSING
                and default_factory() != true_val
            )
        )
        if display_attr:
            kwargs[field_def.name] = true_val

    if hasattr(value, "kwargs"):
        kwargs |= value.kwargs

    return prettyprinter.pretty_call(ctx, cls, **kwargs)


def dfield(default: Any, compare: bool = False, repr: bool = False) -> Any:
    # pylint: disable=redefined-builtin
    return dataclasses.field(default=default, compare=compare, repr=repr)


def default_repr(obj: Any) -> str:
    params = ", ".join([f"{k}={v!r}" for k, v in obj.__dict__.items()])
    return f"{obj.__class__.__qualname__}({params})"


def weighted_coin_flip(prob: float) -> bool:
    """Returns True with probability prob."""
    return random.choices([True, False], [prob, 1 - prob])[0]


formatter = init_prettyprinter()



from __future__ import annotations

import itertools
import random


class RabbitGameState:
    def __init__(self, num_holes: int, verbose: bool) -> None:
        self.holes = [False for _ in range(num_holes)]
        self.rabbit = random.randrange(num_holes)
        self.holes[self.rabbit] = True
        self.guess_count = 0
        self.verbose = verbose
        self.game_over = False
        self.num_holes = num_holes

    def add_guess(self, guess: int | None) -> None:
        if guess is None:
            self.end_game(f"You lost after {self.guess_count} guesses!!")
            return

        self.guess_count += 1
        if self.holes[guess]:
            self.end_game(f"You found the rabbit in {self.guess_count} guesses!!")
        elif self.guess_count > self.num_holes**2:
            self.end_game(f"You lost after {self.guess_count} guesses!!")
        else:
            self.rabbit_hop()

    def end_game(self, message: str) -> None:
        self.game_over = True
        if self.verbose:
            print(message)

    def rabbit_hop(self) -> None:
        self.holes[self.rabbit] = False
        if self.rabbit == 0:
            self.rabbit += 1
        elif self.rabbit == self.num_holes - 1:
            self.rabbit -= 1
        else:
            self.rabbit += random.choice([-1, 1])
        self.holes[self.rabbit] = True


class RabbitFinder3000:
    def __init__(self, num_holes: int) -> None:
        if num_holes == 1:
            self.guesses = iter([0])
        elif 2 <= num_holes <= 3:
            self.guesses = iter([1, 1])
        elif num_holes == 4:
            self.guesses = iter([1, 1, 2, 2, 1])
        else:
            self.guesses = itertools.chain(
                list(range(num_holes)),
                [1] if num_holes % 2 == 0 else [],
                list(range(num_holes)),
            )
        self.num_holes = num_holes

    def guess(self) -> int:
        return next(self.guesses)


def play_game(
    num_holes: int, use_algorithm: bool = False, verbose: bool = False
) -> int:
    rabbit_game = RabbitGameState(num_holes, verbose)
    solver = RabbitFinder3000(num_holes)
    choice_index: int | None
    while not rabbit_game.game_over:
        if use_algorithm:
            choice_index = solver.guess()
        else:
            guess = input(f"What index is the rabbit (0-{num_holes - 1})? ")
            choice_index = int(guess) if guess else None

        rabbit_game.add_guess(choice_index)

    return rabbit_game.guess_count


if __name__ == "__main__":
    play_game(4)



from cs.structures import Graph, V


def connected_components(graph: Graph[V]) -> list[set[V]]:
    """
    This function returns the list of connected components of the given Graph.

    Runtime: O(V + E)
    """
    from cs.algorithms import dfs_traversal

    visited: set[V] = set()
    components_list: list[set[V]] = []
    for v in graph:
        if v not in visited:
            components_list.append(dfs_traversal(graph, v, visited))
    return components_list



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar
from uuid import UUID

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


@dataclass(init=False)
class Heap(Generic[T]):
    def __bool__(self) -> bool:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __contains__(self, item: T) -> bool:
        raise NotImplementedError

    def __ior__(self, other: object) -> None:
        raise NotImplementedError

    def __getitem__(self, value: T | UUID) -> Any:
        raise NotImplementedError

    def enqueue(self, value: T, priority: float = 0) -> T | UUID:
        raise NotImplementedError

    def peek(self) -> tuple[T, float]:
        raise NotImplementedError

    def dequeue(self) -> tuple[T, float]:
        raise NotImplementedError

    def merge(self, other: Heap[T]) -> None:
        raise NotImplementedError



from .binary_search import binary_search, left_right_binary_search, linear_search
from .compression.huffman import huffman_compress, huffman_decompress
from .graph.bellman_ford import bellman_ford_shortest_paths
from .graph.bfs import breadth_first_search
from .graph.connected import connected_components
from .graph.dfs import depth_first_search, dfs_traversal
from .graph.dijkstras import dijkstra_search, dijkstra_shortest_paths
from .graph.edmonds_karp import edmonds_karp_max_flow
from .graph.floyd_warshall import floyd_warshall_shortest_paths
from .graph.ford_fulkerson import bipartite_matching, ford_max_flow
from .graph.hamiltonian_cycle import hamiltonian_cycle
from .graph.johnsons import johnsons_shortest_paths
from .graph.kargers import kargers_min_cut
from .graph.kruskals import kruskals_mst
from .graph.prims import prims_mst
from .graph.toposort import topological_sort
from .optimal_bst import build_optimal_bst
from .quick_select import quick_select
from .sort.bubble_sort import bubble_sort
from .sort.bucket_sort import bucket_sort
from .sort.insertion_sort import insertion_sort
from .sort.merge_sort import merge_sort
from .sort.quick_sort import quick_sort
from .sort.radix_sort import radix_sort
from .sort.selection_sort import selection_sort
from .string.knuth_morris_pratt import kmp_string_match
from .string.lcs import longest_common_subsequence
from .string.sais import build_suffix_array

__all__ = (
    "bellman_ford_shortest_paths",
    "binary_search",
    "bipartite_matching",
    "breadth_first_search",
    "bubble_sort",
    "bucket_sort",
    "build_suffix_array",
    "build_optimal_bst",
    "connected_components",
    "depth_first_search",
    "dfs_traversal",
    "dijkstra_search",
    "dijkstra_shortest_paths",
    "edmonds_karp_max_flow",
    "floyd_warshall_shortest_paths",
    "ford_max_flow",
    "hamiltonian_cycle",
    "huffman_compress",
    "huffman_decompress",
    "insertion_sort",
    "johnsons_shortest_paths",
    "kargers_min_cut",
    "kmp_string_match",
    "kruskals_mst",
    "left_right_binary_search",
    "linear_search",
    "longest_common_subsequence",
    "merge_sort",
    "prims_mst",
    "quick_select",
    "quick_sort",
    "radix_sort",
    "selection_sort",
    "topological_sort",
)



from .rmq import RMQ


class PrecomputedRMQ(RMQ):
    def __init__(self, elems: list[int]) -> None:
        super().__init__(elems)
        self.rmq_table = [[0] * len(elems) for _ in range(len(elems))]

        for i in range(len(elems)):
            self.rmq_table[i][i] = i

        for i in range(1, len(elems)):
            for j in range(i - 1, -1, -1):
                index1 = self.rmq_table[i - 1][j]
                index2 = self.rmq_table[i][j + 1]
                self.rmq_table[i][j] = self.return_smaller_index(index1, index2)

    def rmq(self, low: int, high: int) -> int:
        if low >= high:
            raise RuntimeError("In the range, low must be lower than high.")

        return self.rmq_table[high - 1][low]



from __future__ import annotations

import math
from collections.abc import Iterator


class Bits:
    """
    Personal bits implementation. For all real purposes, use the bitarray library.
    https://docs.python.org/3/library/operator.html
    """

    def __init__(self, val: str = "", *, length: int | None = None) -> None:
        assert length is None or length > 0
        is_negative = len(val) > 0 and val[0] == "-"
        val_check = val[1:] if is_negative else val
        if not all(ch in ("0", "1") for ch in val_check):
            raise RuntimeError(f"{val} is not a valid Bits initial value.")

        self.val = int(val, 2) if val else -1  # -1 == 111111
        self.length = length or len(val)

    def __str__(self) -> str:
        return self.binary_str(self.val, self.length)

    def __repr__(self) -> str:
        return f"Bits({str(self)}, val={self.val}, length={self.length})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            return self.val == other
        if isinstance(other, Bits):
            return str(self) == str(other)
        return NotImplemented

    def __len__(self) -> int:
        return self.length

    def __lshift__(self, other: object) -> Bits:
        if not isinstance(other, int):
            raise TypeError("Bit shifts require an integer shift amount.")
        return Bits(str(self.val * (2**other)), length=self.length)

    def __rshift__(self, other: object) -> Bits:
        if not isinstance(other, int):
            raise TypeError("Bit shifts require an integer shift amount.")
        return Bits(str(self.val // (2**other)), length=self.length)

    def __ilshift__(self, other: object) -> None:
        if not isinstance(other, int):
            raise TypeError("Bit shifts require an integer shift amount.")
        self.val *= 2**other

    def __irshift__(self, other: object) -> None:
        if not isinstance(other, int):
            raise TypeError("Bit shifts require an integer shift amount.")
        self.val //= 2**other

    def __invert__(self) -> Bits:
        """Inverts all bits."""
        return Bits.from_num(~self.val, self.length)

    def __and__(self, other: object) -> Bits:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be intersected with Bits.")
        return Bits.from_num(self.val & other.val, self.length)

    def __iand__(self, other: object) -> None:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be intersected with Bits.")
        self.val &= other.val

    def __or__(self, other: object) -> Bits:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be or'd with Bits.")
        return Bits.from_num(self.val | other.val, self.length)

    def __ior__(self, other: object) -> None:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be or'd with Bits.")
        self.val |= other.val

    def __xor__(self, other: object) -> Bits:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be xor'd with Bits.")
        return Bits.from_num(self.val ^ other.val, self.length)

    def __ixor__(self, other: object) -> None:
        """Intersection of two role sets."""
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be xor'd with Bits.")
        self.val ^= other.val

    def __getitem__(self, index: int) -> int:
        return int(str(self)[index])

    def __setitem__(self, index: int, bit: int) -> None:
        if bit not in (0, 1):
            raise ValueError(f"Bit value must be 0 or 1: {bit}")
        new_val = str(self)
        new_val = new_val[:index] + str(bit) + new_val[index + 1 :]
        self.val = int(new_val, 2) if new_val else -1

    def __iter__(self) -> Iterator[int]:
        for bit in str(self):
            yield int(bit)

    def __add__(self, other: object) -> Bits:
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be added with Bits.")
        length = max(len(self), len(other))
        lhs = [0] * (length - len(self)) + [int(digit) for digit in self]
        rhs = [0] * (length - len(other)) + [int(digit) for digit in other]
        carry = 0
        result = []
        for i in range(1, len(lhs) + 1):
            column = lhs[-i] + rhs[-i] + carry
            result.append(column % 2)
            carry = column // 2
        if carry != 0:
            result.append(carry)
        result.reverse()
        return Bits("".join([str(s) for s in result]))

    def __sub__(self, other: object) -> Bits:
        if not isinstance(other, Bits):
            raise TypeError(f"{other} cannot be subtracted from Bits.")
        length = max(len(self), len(other))
        lhs = [0] * (length - len(self)) + [int(digit) for digit in self]
        rhs = [0] * (length - len(other)) + [int(digit) for digit in other]
        result = []
        for i in range(1, len(lhs) + 1):
            difference = lhs[-i] - rhs[-i]
            if difference >= 0:
                result.append(difference)
                continue
            for j in range(i + 1, length + 1):
                lhs[-j] = (lhs[-j] + 1) % 2
                if lhs[-j] != 1:
                    break
            result.append(difference + 2)
        result.reverse()
        return Bits("".join([str(s) for s in result]))

    @property
    def is_solo(self) -> bool:
        return self.val != 0 and (self.val & (self.val - 1)) == 0

    @property
    def solo(self) -> int:
        """Assumes is_solo is True."""
        if not self.is_solo:
            raise RuntimeError("Does not contain a solo bit.")
        return self.length - int(math.log2(self.val)) - 1

    @classmethod
    def from_num(cls, val: int, length: int) -> Bits:
        return cls(cls.binary_str(val, length), length=length)

    @staticmethod
    def binary_str(val: int, length: int) -> str:
        # return bin(val & (2 ** length - 1))
        coerced_positive_val = val & (2**length - 1)
        return f"{coerced_positive_val:0{length}b}"

    def is_one(self, index: int) -> bool:
        """Returns True if the bit at given index is 1."""
        if index > self.length:
            return False
        return (self.val & (1 << index)) == 0
        # return ((self.val >> index) & 1) == 1

    def set_bit(self, index: int, new_val: bool = True) -> Bits:
        """Mark an index as the given value of its current state."""
        if not 0 <= index < self.length:
            raise RuntimeError("Invalid bit; cannot be set.")

        reversed_index = self.length - index - 1
        if new_val:
            self.val |= 1 << reversed_index
        else:
            self.val &= ~(1 << reversed_index)

        return Bits.from_num(self.val, self.length)

    def flip_index(self, index: int) -> Bits:
        """Mark an index as opposite of its current state."""
        return Bits.from_num(self.val ^ (1 << index), self.length)
        # reversed_index = self.length - index - 1
        # new_val = self.val
        # new_val &= ~(1 << reversed_index)
        # if new_val == self.val:
        #     new_val |= 1 << reversed_index
        # return Bits.from_num(new_val, self.length)



"""
Based on "Skip Lists: A Probabilistic Alternative to Balanced Trees" by William Pugh
https://epaperpress.com/sortsearch/download/skiplist.pdf
"""
from __future__ import annotations

import random
from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import Generic, TypeVar, cast

from cs.util import Comparable

KT = TypeVar("KT", bound=Comparable)
VT = TypeVar("VT")


@dataclass
class SkipList(Generic[KT, VT]):
    key: KT | None = None
    value: VT | None = None
    next: list[SkipList[KT, VT]] = field(default_factory=list)
    p: float = 0.5
    max_level: int = 16
    level: int = 0

    def __str__(self) -> str:
        """
        :return: Visual representation of SkipList

        >>> skip_list = SkipList()
        >>> print(skip_list)
        SkipList(level=0)
        >>> skip_list.insert("Key1", "Value")
        >>> print(skip_list) # doctest: +ELLIPSIS
        SkipList(level=...
        [root]--...
        [Key1]--Key1...
        None    *...
        >>> skip_list.insert("Key2", "OtherValue")
        >>> print(skip_list) # doctest: +ELLIPSIS
        SkipList(level=...
        [root]--...
        [Key1]--Key1...
        [Key2]--Key2...
        None    *...
        """
        items = list(self)
        if len(items) == 0:
            return f"SkipList(level={self.level})"

        key_size = max(len(str(item)) for item in items)
        label_size = max(max((len(str(item)) for item in items), default=4), 4) + 4
        node = self
        forwards = node.next.copy()
        lines = [
            f"[{node.key or 'root'}]".ljust(label_size, "-")
            + f"*{' ' * key_size}" * len(forwards),
            " " * label_size + f"|{' ' * key_size}" * len(forwards),
        ]

        while len(node.next) != 0:
            node = node.next[0]
            lines += [
                f"[{node.key}]".ljust(label_size, "-")
                + " ".join(
                    (str(n.key) if n.key == node.key else "|").ljust(key_size, " ")
                    for n in forwards
                ),
                " " * label_size + f"|{' ' * key_size}" * len(forwards),
            ]
            forwards[: len(node.next)] = node.next

        lines.append("None".ljust(label_size) + f"*{' ' * key_size}" * len(forwards))
        return f"______SkipList(level={self.level})______\n" + "\n".join(lines)

    def __iter__(self) -> Iterator[KT]:
        node = self
        while len(node.next) != 0:
            yield cast(KT, node.next[0].key)
            node = node.next[0]

    def __getitem__(self, key: KT) -> VT:
        node, _ = self._locate_node(key)
        if node is None or node.value is None:
            raise KeyError
        return node.value

    def __contains__(self, key: KT) -> bool:
        """
        :param key: Search key.
        :return: Value associated with given key
            or None if given key is not present.

        >>> skip_list = SkipList()
        >>> 2 in skip_list
        False
        >>> skip_list.insert(2, "Two")
        >>> 2 in skip_list
        True
        >>> skip_list[2]
        'Two'
        >>> skip_list.insert(2, "Three")
        >>> skip_list[2]
        'Three'
        """
        node, _ = self._locate_node(key)
        return node is not None

    def random_level(self) -> int:
        """
        :return: Random level from [1, self.max_level] interval.
                 Higher values are less likely.
        """
        level = 1
        while random.random() < self.p and level < self.max_level:
            level += 1
        return level

    def remove(self, key: KT) -> None:
        """
        :param key: Key to remove from list.

        >>> skip_list = SkipList()
        >>> skip_list.insert(2, "Two")
        >>> skip_list.insert(1, "One")
        >>> skip_list.insert(3, "Three")
        >>> list(skip_list)
        [1, 2, 3]
        >>> skip_list.remove(2)
        >>> list(skip_list)
        [1, 3]
        """
        node, update_vector = self._locate_node(key)
        if node is None:
            raise KeyError(f"List does not contain: {key}")

        for i, update_node in enumerate(update_vector):
            # Remove or replace all references to removed node.
            if len(update_node.next) > i and update_node.next[i].key == key:
                if len(node.next) > i:
                    update_node.next[i] = node.next[i]
                else:
                    update_node.next = update_node.next[:i]

    def insert(self, key: KT, value: VT) -> None:
        """
        :param key: Key to insert.
        :param value: Value associated with given key.

        >>> skip_list = SkipList()
        >>> skip_list.insert(2, "Two")
        >>> skip_list[2]
        'Two'
        >>> list(skip_list)
        [2]
        """
        node, update_vector = self._locate_node(key)
        if node is not None:
            node.value = value
            return

        level = self.random_level()
        if level > self.level:
            # After level increase we have to add additional nodes to head.
            for _ in range(self.level - 1, level):
                update_vector.append(self)
            self.level = level

        new_node = SkipList(key, value)
        for i, update_node in enumerate(update_vector[:level]):
            # Change references to pass through new node.
            if len(update_node.next) > i:
                new_node.next.append(update_node.next[i])
            if len(update_node.next) < i + 1:
                update_node.next.append(new_node)
            else:
                update_node.next[i] = new_node

    def _locate_node(
        self, key: KT
    ) -> tuple[SkipList[KT, VT] | None, list[SkipList[KT, VT]]]:
        """
        :param key: Searched key,
        :return: Tuple with searched node (or None if given key is not present)
                 and list of nodes that refer (if key is present) of should refer to
                 given node.
        """
        # SkipListNodes with refer or should refer to output node
        update_vector = []
        node = self

        for i in reversed(range(self.level)):
            # i < node.level - When node level is lesser than `i` decrement `i`.
            # node.next[i].key < key - Jumping to node with key value higher
            #                             or equal to searched key would result
            #                             in skipping searched key.
            while (
                i < len(node.next) and node.next[i].key < key  # type: ignore[operator]
            ):
                node = node.next[i]
            # Each leftmost node (relative to searched node) will potentially have to
            # be updated.
            update_vector.append(node)

        update_vector.reverse()  # Note that we were inserting values in reverse order.

        # len(node.next) != 0 - If current node doesn't contain any further
        #                          references then searched key is not present.
        # node.next[0].key == key - Next node key should be equal to search key
        #                              if key is present.
        if len(node.next) != 0 and node.next[0].key == key:
            return node.next[0], update_vector
        return None, update_vector



from __future__ import annotations

import math

from .hash_table import KT, VT, HashTable, TableEntry


class Cuckoo(HashTable[KT, VT]):
    """
    A hash system with worst-case constant-time lookup and deletion, and amortized
    expected O(1) insertion.

    Cuckoo hashing works by maintaining two arrays and two universal hash functions f
    and g. When an element x is inserted, the value f(x) is computed and the entry is
    stored in that index in the first array. If that spot was initially empty, we are
    done. Otherwise, the element y that was already there is "kicked out." We then
    compute g(y) and store element y at position g(y) in the second array, which may in
    turn kick out another element, which will be stored in the first array. This process
    repeats until either a loop is detected (in which case we pick a new hash function
    and rehash), or all elements finally come to rest.
    """

    def __init__(self, num_buckets: int, load_factor: float = 0.4) -> None:
        super().__init__(num_buckets, load_factor)
        self.num_buckets //= 2
        if self.num_buckets <= 1:
            self.num_buckets += 1

        self.table_1: list[TableEntry[KT, VT] | None] = [None] * self.num_buckets
        self.table_2 = list(self.table_1)
        self.hash_1 = self.generate_hash_function()
        self.hash_2 = self.generate_hash_function()
        self.rehashing_depth_limit = int(6 * math.log(num_buckets * 2))

    def __str__(self) -> str:
        max_width = 20
        result = ""
        for i in range(self.num_buckets):
            table_row_1 = f"{i}  |  {self.table_1[i]}"
            table_row_2 = f"{i}  |  {self.table_2[i]}"
            result += f"{table_row_1:<{max_width}} {table_row_2}\n"
        return result

    def insert(self, key: KT, value: VT) -> None:
        self.validate_key(key)
        # Replace existing key
        if (entry := self._find_key(key)) is not None:
            entry.value = value
            return

        # Resize table if size exceeds capacity.
        if self.num_elems >= self.load_factor * self.capacity:
            self.capacity *= 2
            self.num_buckets *= 2
            self._rehash()

        self._insert(key, value)
        self.num_elems += 1

    def remove(self, key: KT) -> None:
        self.validate_key(key)
        self.num_elems -= 1
        bucket = self.hash_1(key) % self.num_buckets
        entry = self.table_1[bucket]
        if entry is not None and entry.key == key:
            self.table_1[bucket] = None
            return

        bucket = self.hash_2(key) % self.num_buckets
        entry = self.table_2[bucket]
        if entry is not None and entry.key == key:
            self.table_2[bucket] = None

    def _insert(self, key: KT, value: VT) -> None:
        use_table_1 = True
        curr: TableEntry[KT, VT] | None = TableEntry(key, value)
        # Try inserting key, swapping items up to the given depth limit.
        for _ in range(self.rehashing_depth_limit):
            if curr is None:
                raise RuntimeError

            if use_table_1:
                bucket = self.hash_1(curr.key) % self.num_buckets
                if self.table_1[bucket] is None:
                    self.table_1[bucket] = curr
                    return
                curr, self.table_1[bucket] = self.table_1[bucket], curr

            else:
                bucket = self.hash_2(curr.key) % self.num_buckets
                if self.table_2[bucket] is None:
                    self.table_2[bucket] = curr
                    return
                curr, self.table_2[bucket] = self.table_2[bucket], curr

            use_table_1 = not use_table_1

        self._rehash()

    def _rehash(self) -> None:
        # Rehash using a new hash function and recurse to try insert again.
        self.hash_1 = self.generate_hash_function()
        self.hash_2 = self.generate_hash_function()

        # Copy all old elements to a temp table
        table = [item for item in self.table_1 + self.table_2 if item is not None]

        # Clear both tables
        self.table_1 = [None] * self.num_buckets
        self.table_2 = list(self.table_1)

        # Reinsert all elements
        for item in table:
            self._insert(item.key, item.value)

    def _find_key(self, key: KT) -> TableEntry[KT, VT] | None:
        bucket = self.hash_1(key) % self.num_buckets
        entry = self.table_1[bucket]
        if entry is not None and entry.key == key:
            return entry

        bucket = self.hash_2(key) % self.num_buckets
        entry = self.table_2[bucket]
        if entry is not None and entry.key == key:
            return entry

        return None



from .ado.ado_finite_metric import ApproxFiniteMetricOracle
from .ado.ado_graph import ApproxDistanceOracle
from .bloom_filter import BloomFilter
from .disjoint_set import DisjointSet
from .graph import DirectedGraph, Edge, Graph, Node, UndirectedGraph, V
from .hash_table.cuckoo import Cuckoo
from .hash_table.hash_table import HashTable
from .hash_table.linear_probing import LinearProbing
from .hash_table.robinhood import RobinHood
from .heap.binary_heap import BinaryHeap
from .heap.binomial_heap import BinomialHeap
from .heap.fibonacci_heap import FibonacciHeap
from .heap.heap import Heap
from .linked_list.doubly_linked_list import DoublyLinkedList
from .linked_list.linked_list import LinkedList
from .linked_list.skip_list import SkipList
from .lru_cache import LRUCache, lru_cache
from .queue import Queue
from .rmq.fischer_heun_rmq import FischerHeunRMQ
from .rmq.hybrid_rmq import HybridRMQ
from .rmq.precomputed_rmq import PrecomputedRMQ
from .rmq.rmq import RMQ
from .rmq.sparse_table_rmq import SparseTableRMQ
from .stack import Stack
from .suffix_array import SuffixArray
from .tree.binary_search_tree import BinarySearchTree, BinaryTreeNode
from .tree.red_black_tree import RedBlackTree, RedBlackTreeNode
from .tree.tree import Tree, TreeNode
from .trie import Trie

__all__ = (
    "ApproxDistanceOracle",
    "ApproxFiniteMetricOracle",
    "BinaryHeap",
    "BinarySearchTree",
    "BinaryTreeNode",
    "BinomialHeap",
    "BloomFilter",
    "Cuckoo",
    "DirectedGraph",
    "DisjointSet",
    "DoublyLinkedList",
    "Edge",
    "FibonacciHeap",
    "FischerHeunRMQ",
    "Graph",
    "HashTable",
    "Heap",
    "HybridRMQ",
    "LRUCache",
    "LinearProbing",
    "LinkedList",
    "Node",
    "PrecomputedRMQ",
    "Queue",
    "RMQ",
    "RedBlackTree",
    "RedBlackTreeNode",
    "RobinHood",
    "SkipList",
    "SparseTableRMQ",
    "Stack",
    "SuffixArray",
    "Tree",
    "TreeNode",
    "Trie",
    "UndirectedGraph",
    "V",
    "lru_cache",
)



def celsius_to_fahrenheit(celsius: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Celsius to Fahrenheit and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    """
    return round(celsius * 9 / 5 + 32, ndigits)


def celsius_to_kelvin(celsius: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Celsius to Kelvin and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    """
    return round(celsius + 273.15, ndigits)


def celsius_to_rankine(celsius: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Celsius to Rankine and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    """
    return round(celsius * 9 / 5 + 491.67, ndigits)


def fahrenheit_to_celsius(fahrenheit: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Fahrenheit to Celsius and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    """
    return round((fahrenheit - 32) * 5 / 9, ndigits)


def fahrenheit_to_kelvin(fahrenheit: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Fahrenheit to Kelvin and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    """
    return round(((fahrenheit - 32) * 5 / 9) + 273.15, ndigits)


def fahrenheit_to_rankine(fahrenheit: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Fahrenheit to Rankine and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    """
    return round(fahrenheit + 459.67, ndigits)


def kelvin_to_celsius(kelvin: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Kelvin to Celsius and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    """
    return round(kelvin - 273.15, ndigits)


def kelvin_to_fahrenheit(kelvin: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Kelvin to Fahrenheit and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    """
    return round((kelvin - 273.15) * 9 / 5 + 32, ndigits)


def kelvin_to_rankine(kelvin: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Kelvin to Rankine and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    """
    return round(kelvin * 9 / 5, ndigits)


def rankine_to_celsius(rankine: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Rankine to Celsius and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    Wikipedia reference: https://en.wikipedia.org/wiki/Celsius
    """
    return round((rankine - 491.67) * 5 / 9, ndigits)


def rankine_to_fahrenheit(rankine: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Rankine to Fahrenheit and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    Wikipedia reference: https://en.wikipedia.org/wiki/Fahrenheit
    """
    return round(rankine - 459.67, ndigits)


def rankine_to_kelvin(rankine: float, ndigits: int = 2) -> float:
    """
    Convert a given value from Rankine to Kelvin and round it to 2 decimal places.
    Wikipedia reference: https://en.wikipedia.org/wiki/Rankine_scale
    Wikipedia reference: https://en.wikipedia.org/wiki/Kelvin
    """
    return round(rankine * 5 / 9, ndigits)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from .hash_table import KT, VT, HashTable, TableEntry


@dataclass(slots=True)
class RobinHoodEntry(TableEntry[KT, VT]):
    dist: int = -1

    def __repr__(self) -> str:
        return f"({self.value}, {self.dist})"


class RobinHood(HashTable[KT, VT]):
    def __init__(self, num_buckets: int, load_factor: float = 0.4) -> None:
        super().__init__(num_buckets, load_factor)
        self.table: list[RobinHoodEntry[KT, VT] | None] = [None] * num_buckets
        self.hash: Callable[[KT], int] = lambda x: hash(x) % num_buckets

    def __str__(self) -> str:
        result = ""
        for i in range(self.num_buckets):
            entry = self.table[i]
            result += f"{i}  |  {None if entry is None else entry.value}\n"
        return result

    def insert(self, key: KT, value: VT) -> None:
        self.validate_key(key)
        # Replace existing key
        if (entry := self._find_key(key)) is not None:
            entry.value = value
            return

        # Resize table if size exceeds capacity.
        if self.num_elems >= self.load_factor * self.capacity:
            self.capacity *= 2
            self.num_buckets *= 2
            self.num_elems = 0
            table = list(self.table)
            self.table = [None] * self.num_buckets
            for entry in table:
                if entry is not None:
                    self.insert(entry.key, entry.value)

        bucket = self.hash(key) % self.num_buckets
        curr: RobinHoodEntry[KT, VT] | None = RobinHoodEntry(key, value)
        curr_dist = 0
        while curr is not None and (
            (entry := self.table[bucket]) is None or curr.key != entry.key
        ):
            if entry is None or entry.dist < curr_dist:
                self.table[bucket] = RobinHoodEntry(curr.key, curr.value, curr_dist)
                curr, curr_dist = entry, -1 if entry is None else entry.dist
            curr_dist += 1
            bucket = (bucket + 1) % self.num_buckets
        self.num_elems += 1

    def remove(self, key: KT) -> None:
        self.validate_key(key)
        bucket = self.hash(key) % self.num_buckets
        for i in range(self.num_buckets):
            index = (bucket + i) % self.num_buckets
            entry = self.table[index]
            if entry is None:
                break
            if entry.key == key:
                self._backward_shift(index)
                self.num_elems -= 1
                return
        raise KeyError

    def _backward_shift(self, i: int) -> None:
        j = (i + 1) % self.num_buckets
        entry = self.table[j]
        while entry is not None and entry.dist > 0:
            entry.dist -= 1
            self.table[i], entry = entry, self.table[i]
            i, j = j, (i + 1) % self.num_buckets
        self.table[i] = None

    def _find_key(self, key: KT) -> RobinHoodEntry[KT, VT] | None:
        bucket = self.hash(key) % self.num_buckets
        for i in range(self.num_buckets):
            index = (bucket + i) % self.num_buckets
            entry = self.table[index]
            if entry is not None and entry.key == key:
                return entry
            if entry is None or entry.dist < i:
                break
        return None



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(repr=False, slots=True)
class DoublyLinkedListNode(Generic[T]):
    data: T
    prev: DoublyLinkedListNode[T] | None = None
    next: DoublyLinkedListNode[T] | None = None

    def __repr__(self) -> str:
        return f"({self.data}) -> {self.next}"


@dataclass(init=False)
class DoublyLinkedList(Generic[T]):
    """
    - A linked list is similar to an array, it holds values. However, links
        in a linked list do not have indexes.
    - This is an example of a double ended, doubly linked list (or a deque).
    - Each link references the next link and the prev one.
    - A Doubly Linked List (DLL) contains an extra pointer, typically called
        prev pointer, together with next pointer and data which are
        there in singly linked list.
    - Advantages over SLL - It can be traversed in both forward and
        backward direction. Delete operation is more efficient.
    """

    head: DoublyLinkedListNode[T] | None

    def __init__(self) -> None:
        self.head: DoublyLinkedListNode[T] | None = None
        self.tail: DoublyLinkedListNode[T] | None = None
        self.size = 0

    def __len__(self) -> int:
        return self.size

    def __iter__(self) -> Iterator[DoublyLinkedListNode[T]]:
        node = self.head
        while node is not None:
            yield node
            node = node.next

    def __bool__(self) -> bool:
        return self.head is not None

    def __contains__(self, item: T) -> bool:
        curr = self.head
        while curr is not None:
            if curr.data == item:
                return True
            curr = curr.next
        return False

    def __getitem__(self, index: int) -> T:
        current = self.head
        if current is None:
            raise IndexError("List is empty")
        for _ in range(index):
            if current.next is None:
                raise IndexError("Index out of range.")
            current = current.next
        return current.data

    def __setitem__(self, index: int, data: T) -> None:
        current = self.head
        if current is None:
            raise IndexError("List is empty")
        for _ in range(index):
            if current.next is None:
                raise IndexError("Index out of range.")
            current = current.next
        current.data = data

    @classmethod
    def from_list(cls, lst: list[T]) -> DoublyLinkedList[T]:
        linked_lst = cls()
        for item in reversed(lst):
            linked_lst.insert_at_head(item)
        return linked_lst

    def insert_at_head(self, data: T) -> None:
        new_node = DoublyLinkedListNode(data)
        self.size += 1
        if self.head is None or self.tail is None:
            self.tail = new_node
            self.head = new_node
        else:
            self.head.prev = new_node
            new_node.next = self.head
            self.head = new_node

    def insert_at_tail(self, data: T) -> None:
        new_node = DoublyLinkedListNode(data)
        self.size += 1
        if self.head is None or self.tail is None:
            self.tail = new_node
            self.head = new_node
        else:
            self.tail.next = new_node
            new_node.prev = self.tail
            self.tail = new_node

    def pop_head(self) -> T:
        if self.head is None:
            raise Exception("List is empty")

        self.size -= 1
        head_data = self.head.data
        if self.head.next:
            self.head = self.head.next
            self.head.prev = None
        else:  # If there is no next prev node
            self.head = None
            self.tail = None
        return head_data

    def pop_tail(self) -> T:
        if self.head is None or self.tail is None:
            raise Exception("List is empty")

        self.size -= 1
        tail_data = self.tail.data
        if self.tail.prev is not None:
            self.tail = self.tail.prev
            self.tail.next = None
        else:  # if there is no prev node
            self.head = None
            self.tail = None
        return tail_data

    def remove(self, data: T) -> None:
        if self.head is None or self.tail is None:
            raise Exception("List is empty")

        current = self.head
        while current.data != data:  # Find the position to delete
            if current.next is not None:
                current = current.next
            else:  # We have reached the end an no value matches
                raise Exception("List does not contain value.")

        if current == self.head:
            _ = self.pop_head()
        elif current == self.tail:
            _ = self.pop_tail()
        else:
            self.size -= 1
            if current.prev is None or current.next is None:
                raise ValueError
            current.prev.next = current.next
            current.next.prev = current.prev



"""
ADO for Finite Metric Spaces
This code is largely untested. You can look in tests/ to see some examples, or
run them using `pytest -vv`.
It seems to work fine, but take that with a grain of salt.
"""
import math

from cs.util import weighted_coin_flip

Vertex = int


class ApproxFiniteMetricOracle:
    def __init__(self, distance_matrix: list[list[float]], k: int = 2) -> None:
        """
        Preprocessing step.
        Initialize k+1 sets of vertices with decreasing sizes. (i-centers)
        a_i_v_distances:
            stores distances from any vertex to one of the i-centers.
            (the closest one to the vertex)
        self.p:
            p[i][v] is the vertex in i-center that is nearest to v
        self.B:
            Bunches!
            B[v] contains the union of all sets B_i.
            B_i contains all vertices in A[i] that are strictly closer to v
            than all vertices in A[i-1]
                The partial unions of B_{i}s are balls in increasing diameter,
                that contain vertices with distances up to the first
                vertex of the next level.
        Notes:
           for fixed v, the distance is weakly increasing with i
           for all v, a_i_v_distances[0][v] = 0 and p[0][v] = v
        """
        n = len(distance_matrix)
        self.distance_matrix = distance_matrix
        self.n = n
        self.k = k

        V = set(range(n))
        A = [V]
        for i in range(1, k):
            prob = n ** (-1 / k)
            A_i = {v for v in A[i - 1] if weighted_coin_flip(prob)}
            A.append(A_i)
        A.append(set())

        # Find minimum distances from each vertex to each other set.
        a_i_v_distances: list[dict[int, float]] = [{} for _ in range(k + 1)]
        self.p: list[dict[int, Vertex]] = [{} for _ in range(k + 1)]
        self.B: dict[int, set[int]] = {}

        for v in V:
            for i in range(k):
                min_dist, w = self.find_closest_vertex(A[i], v)
                a_i_v_distances[i][v] = min_dist
                self.p[i][v] = w
            a_i_v_distances[k][v] = math.inf

            self.B[v] = set()
            for i in range(k):
                self.B[v] |= {
                    w
                    for w in A[i] - A[i + 1]
                    if self.distance_fn(w, v) < a_i_v_distances[i + 1][v]
                }

        self.hash_table = {}
        for v, b_set in self.B.items():
            for w in b_set:
                self.hash_table[(w, v)] = self.distance_fn(w, v)

    def find_closest_vertex(self, A_i: set[Vertex], v: Vertex) -> tuple[float, Vertex]:
        """
        The result of the below code is that:
            distances[(A[i], v)] = min([ distances[(w, v)] for w in A[i] ])
            p[(i, v)] = w
        """
        min_dist = math.inf
        for w in A_i:  # iterate over vertices
            computed_dist = self.distance_fn(w, v)
            if computed_dist < min_dist:
                min_dist = computed_dist
                closest_w = w
        return min_dist, closest_w

    def query(self, u: Vertex, v: Vertex) -> float:
        w = u
        i = 0
        while w not in self.B[v]:
            i += 1
            u, v = v, u
            w = self.p[i][u]
        return self.hash_table[(w, u)] + self.hash_table[(w, v)]

    def distance_fn(self, u: Vertex, v: Vertex) -> float:
        return self.distance_matrix[u][v]



from collections import defaultdict

from cs.structures.graph import Graph
from cs.structures.heap.fibonacci_heap import FibonacciHeap
from cs.util import weighted_coin_flip


class ApproxDistanceOracle:
    def __init__(self, graph: Graph[int], k: int = 2) -> None:
        """
        Preprocessing step.
        Note that:
            self.p represents witnesses
            self.a_i_v_distances represents delta(A_i, v)
        """
        self.graph = graph
        self.n = len(graph)

        # Initialize k+1 sets of vertices with decreasing sizes. (i-centers)
        self.A: list[set[int]] = [set()] * (k + 1)
        self.A[0] = set(self.graph)
        self.A[k] = set()
        for i in range(1, k):  # for i = 1 to k - 1
            prob = self.n ** (-1 / k)
            self.A[i] = {x for x in self.A[i - 1] if weighted_coin_flip(prob)}

        self.a_i_v_distances: list[list[float]] = [[0] * self.n for _ in range(k + 1)]
        self.p: list[list[int]] = [[0] * self.n for _ in range(k + 1)]

        # Initialize a_i_v_distances of A_k to INFINITY
        self.a_i_v_distances[k] = [Graph.INFINITY] * self.n
        self.p[k] = [0] * self.n

        # Initialize empty bunches
        self.B: list[set[int]] = [{v} for v in self.graph]

        # Initialize table of calculated distances
        self.distances: dict[tuple[int, int], float] = defaultdict(
            lambda: Graph.INFINITY
        )
        for v in self.graph:
            self.distances[(v, v)] = 0

        # for i = k - 1 down to 0
        for i in range(k - 1, -1, -1):
            # compute delta(A_i, v) for each v in V
            self.compute_delta_a_i_v(i)

            # compute distances and bunches
            self.compute_vertex_distances(i)

    def query(self, u: int, v: int) -> float:
        w = u
        i = 0
        while w not in self.B[v]:
            i += 1
            u, v = v, u
            w = self.p[i][u]
        return self.distances[(w, u)] + self.distances[(w, v)]

    def compute_delta_a_i_v(self, i: int) -> None:
        """Variant on Dijkstra's that tracks witnesses."""
        q = UIntPQueue()
        self.a_i_v_distances[i] = [Graph.INFINITY] * self.n
        self.p[i] = [Graph.INFINITY] * self.n  # type: ignore[list-item]
        for w in self.A[i]:
            self.a_i_v_distances[i][w] = 0
            self.p[i][w] = w
            # Instead of adding and later removing a new source vertex, just
            # enqueue everything in A_i
            q.enqueue(w, 0)
        while q:
            w = q.dequeue()
            for v in self.graph[w]:
                prev = self.a_i_v_distances[i][v]
                nxt = self.a_i_v_distances[i][w] + self.graph[w][v].weight
                if nxt < prev:
                    self.a_i_v_distances[i][v] = nxt
                    self.p[i][v] = self.p[i][w]
                    q.enqueue(v, nxt)

    def compute_vertex_distances(self, i: int) -> None:
        """
        A modified version of Dijkstra's algorithm that only updates delta(c, v)
        if the new estimate of delta(c, v) is strictly smaller than delta(A_(i+1), v).
        """
        q = UIntPQueue()

        # Run Dijkstra's algorithm from each i-center
        for c in self.A[i]:
            q.enqueue(c, 0)
            while q:
                w = q.dequeue()
                for v in self.graph[w]:
                    nxt = self.distances[(c, w)] + self.graph[w][v].weight
                    # Only store the distance if the i-center c is closer to v
                    # than everything in A_(i+1)
                    if nxt < self.a_i_v_distances[i + 1][v]:
                        prev = self.distances[(c, v)]
                        if nxt < prev:
                            self.distances[(c, v)] = nxt
                            self.B[v].add(c)
                            q.enqueue(v, nxt)


class UIntPQueue:
    """
    A priority queue that does not allow repeats. When a repeat value is
    enqueued, it is updated with the smaller priority. The queue only allows
    nonnegative integers.
    """

    def __init__(self) -> None:
        self.fheap = FibonacciHeap[int]()

    def __len__(self) -> int:
        return len(self.fheap)

    def enqueue(self, value: int, priority: float) -> None:
        if value not in self.fheap:
            self.fheap.enqueue(value, priority)
        elif priority < self.fheap[value].value:
            self.fheap.decrease_key(value, priority)

    def dequeue(self) -> int:
        value, _ = self.fheap.dequeue()
        return value



"""
Convert International System of Units (SI) and Binary prefixes
"""
from enum import IntEnum, unique


@unique
class SI_Unit(IntEnum):
    yotta = 24
    zetta = 21
    exa = 18
    peta = 15
    tera = 12
    giga = 9
    mega = 6
    kilo = 3
    hecto = 2
    deca = 1
    deci = -1
    centi = -2
    milli = -3
    micro = -6
    nano = -9
    pico = -12
    femto = -15
    atto = -18
    zepto = -21
    yocto = -24


@unique
class Binary_Unit(IntEnum):
    yotta = 8
    zetta = 7
    exa = 6
    peta = 5
    tera = 4
    giga = 3
    mega = 2
    kilo = 1


def convert_si_prefix(
    known_amount: float, from_prefix: SI_Unit, to_prefix: SI_Unit
) -> float:
    difference: float = from_prefix - to_prefix
    return known_amount * float(10**difference)


def convert_binary_prefix(
    known_amount: float, from_prefix: Binary_Unit, to_prefix: Binary_Unit
) -> float:
    difference: float = from_prefix.value - to_prefix.value
    return known_amount * float(2 ** (difference * 10))



from collections import Counter
from dataclasses import dataclass
from enum import Enum, unique


@unique
class SuffixType(Enum):
    S, L = "S", "L"


@dataclass(slots=True)
class LMSBlock:
    start: int
    end: int
    block_num: int = -1


def build_suffix_array_naive(source: str) -> list[int]:
    """
    A naive, slow suffix-array construction algorithm.
    Construct a list of suffixes of the source string, then calculate the start
    offset of each suffix, storing them in sorted order into the suffix array.

    Runtime: O(n log n)
    """
    n = len(source)
    suffixes = sorted(source[offset:] for offset in range(n + 1))
    return [n - len(suffix) for suffix in suffixes]


def build_suffix_array(orig_text: str) -> list[int]:
    """
    SAIS algorithm to form a Suffix Array.

    Runtime: O(n)
    """

    def _sais(text: list[int]) -> list[int]:
        """Helper function required to handle the recursive call."""

        # Handle empty string edge case
        if not orig_text:
            return [0]

        # Step 1: Annotate all suffixes as L/S and get LMS suffixes.
        suffix_marks, lms_suffixes = get_suffix_annotations(text)

        # Step 2: Run Induced sorting on the suffix array.
        suffix_arr = induced_sort(text, suffix_marks, lms_suffixes)

        # Step 3: Get the reduced string.
        reduced_str, should_recurse = get_reduced_string(text, suffix_arr, lms_suffixes)

        # Step 4: Sort the reduced string by computing its suffix array!
        if should_recurse:
            reduced_suffix_arr = _sais(reduced_str)
        else:
            reduced_suffix_arr = [-1] * len(reduced_str)
            for i, item in enumerate(reduced_str):
                reduced_suffix_arr[item] = i

        sorted_lms_suffixes = reorder_lms_substrings(lms_suffixes, reduced_suffix_arr)

        # Step 5: Final induced sorting pass.
        return induced_sort(text, suffix_marks, sorted_lms_suffixes)

    # Step 0: Convert text to rank array.
    return _sais(to_rank_array(orig_text))


def create_lms_blocks(
    suffix_arr: list[int], lms_suffixes: list[int]
) -> tuple[list[LMSBlock], dict[int, int]]:
    """
    This function fetches all LMS blocks, which span from one LMS suffix to another.
    """
    # Stores the index in the original lms_suffix list.
    is_lms_suffix = [-1] * len(suffix_arr)
    for index, suffix in enumerate(lms_suffixes):
        is_lms_suffix[suffix] = index
    sorted_lms_suffixes = [item for item in suffix_arr if is_lms_suffix[item] != -1]

    lms_blocks = []
    block_mapping = {}
    for i, curr_suffix in enumerate(sorted_lms_suffixes):
        search = is_lms_suffix[curr_suffix] + 1
        next_suffix = (
            curr_suffix if search == len(lms_suffixes) else lms_suffixes[search]
        )
        lms_blocks.append(LMSBlock(curr_suffix, next_suffix))
        block_mapping[curr_suffix] = i
    return lms_blocks, block_mapping


def assign_block_numbers(text: list[int], lms_blocks: list[LMSBlock]) -> bool:
    """
    Iterate over the sorted LMS blocks and assign them block numbers in place. If any
    two have the same block number, make a note to recurse later to sort the
    LMS suffixes. We assume that lmsBlocks is in sorted order by appearance in text.
    """
    lms_blocks[0].block_num = 0
    curr_block = 1
    should_recurse = False
    for i in range(1, len(lms_blocks)):
        block = lms_blocks[i]
        prev_block = lms_blocks[i - 1]
        if (
            text[block.start : block.end + 1]
            == text[prev_block.start : prev_block.end + 1]
        ):
            block.block_num = prev_block.block_num
            should_recurse = True
        else:
            block.block_num = curr_block
            curr_block += 1
    return should_recurse


def get_reduced_string(
    text: list[int], suffix_arr: list[int], lms_suffixes: list[int]
) -> tuple[list[int], bool]:
    """
    >>> suffix_arr = [10, 0, 1, 2, 3, 4, 6, 8, 5, 7, 9]
    >>> lms_suffixes = [6, 8, 10]
    >>> get_reduced_string("AAAAACACAG", suffix_arr, lms_suffixes)
    ([1, 2, 0], False)
    """
    lms_blocks, block_mapping = create_lms_blocks(suffix_arr, lms_suffixes)
    should_recurse = assign_block_numbers(text, lms_blocks)

    reduced_str = []
    for suffix in lms_suffixes:
        block_index = block_mapping[suffix]
        reduced_str.append(lms_blocks[block_index].block_num)
    return reduced_str, should_recurse


def reorder_lms_substrings(
    lms_suffixes: list[int], reduced_str: list[int]
) -> list[int]:
    """
    >>> reorder_lms_substrings([2, 6, 8, 11, 13, 16, 20], [6, 5, 3, 1, 0, 4, 2])
    [20, 16, 11, 6, 2, 13, 8]
    """
    return [lms_suffixes[reduced_str[i]] for i in range(len(reduced_str))]


def induced_sort(
    text: list[int], suffix_marks: list[SuffixType], lms_suffixes: list[int]
) -> list[int]:
    """
    Runs induced sort on the text and returns a suffix array.
    """
    alphabet_size = max(text)
    suffix_arr = [-1] * len(text)

    histogram = Counter(text)
    bucket_starts = [0] * (alphabet_size + 1)  # text -> start ind
    bucket_ends = list(bucket_starts)

    for i in range(1, alphabet_size + 1):
        bucket_starts[i] = bucket_starts[i - 1] + histogram[i - 1]
        bucket_ends[i] = bucket_ends[i - 1] + histogram[i]

    orig_bucket_ends = list(bucket_ends)

    # Make backward pass and insert all LMS suffixes into their
    # proper places at the bucket ends.
    for lms_suffix in reversed(lms_suffixes):
        bucket = bucket_ends[text[lms_suffix]]
        suffix_arr[bucket] = lms_suffix
        bucket_ends[text[lms_suffix]] -= 1

    # Make forward pass and insert all L-type suffixes
    # into the proper places at the bucket starts.
    for key in suffix_arr:
        if key > 0:
            ind = key - 1
            if suffix_marks[ind] == SuffixType.L:
                bucket = bucket_starts[text[ind]]
                suffix_arr[bucket] = ind
                bucket_starts[text[ind]] += 1

    # Make final backward pass and reinsert all S-type suffixes into the proper
    # places at the bucket ends, overwriting any previously placed suffixes.
    bucket_ends = orig_bucket_ends
    for i in range(len(suffix_arr) - 1, 0, -1):
        key = suffix_arr[i]
        if key > 0:
            ind = key - 1
            if suffix_marks[ind] is SuffixType.S:
                bucket = bucket_ends[text[ind]]
                suffix_arr[bucket] = ind
                bucket_ends[text[ind]] -= 1

    return suffix_arr


def get_suffix_annotations(text: list[int]) -> tuple[list[SuffixType], list[int]]:
    """
    Returns all LMS suffixes, as well as the L/S markings for each index.
    >>> marks, lms = get_suffix_annotations("ACGTGCCTAGCCTACCGTGCC$")
    >>> [x.value for x in marks]  # doctest: +NORMALIZE_WHITESPACE
    ['S', 'S', 'S', 'L', 'L', 'S', 'S', 'L', 'S', 'L', 'S',
     'S', 'L', 'S', 'S', 'S', 'S', 'L', 'L', 'L', 'L', 'S']
    >>> lms
    [5, 8, 10, 13, 21]
    """
    lms_suffixes: list[int] = []
    suffix_marks = [SuffixType.S] * len(text)
    for k in range(len(text) - 1, 0, -1):
        if (
            text[k - 1] > text[k]
            or text[k - 1] == text[k]
            and suffix_marks[k] is SuffixType.L
        ):
            suffix_marks[k - 1] = SuffixType.L
            if suffix_marks[k] is SuffixType.S:
                lms_suffixes.append(k)

    lms_suffixes.reverse()
    return suffix_marks, lms_suffixes


def to_rank_array(text: str) -> list[int]:
    """
    Gets rank array representation for a text. Appends a sentinel $ at the end.
    >>> to_rank_array("abracadabra")
    [1, 2, 5, 1, 3, 1, 4, 1, 2, 5, 1, 0]
    >>> to_rank_array("ACGTGCCTAGCCTACCGTGCC")
    [1, 2, 3, 4, 3, 2, 2, 4, 1, 3, 2, 2, 4, 1, 2, 2, 3, 4, 3, 2, 2, 0]
    """
    chars = sorted(set(text))
    char_map: dict[str, int] = {}
    for ch in chars:
        char_map[ch] = len(char_map) + 1
    result: list[int] = []
    for ch in text:
        result.append(char_map[ch])
    # End string terminator
    result.append(0)
    return result


def rank_text_to_str(char_map: dict[str, int], rank_text: list[int]) -> str:
    """Util function to convert rank text back into strings."""
    result = ""
    reversed_char_map = {}
    reversed_char_map[0] = "$"
    for ch in char_map:
        reversed_char_map[char_map[ch]] = ch
    for num in rank_text:
        result += reversed_char_map[num]
    return result



from __future__ import annotations

import heapq

from cs.structures import FibonacciHeap, Graph, V


def dijkstra_search(graph: Graph[V], start: V, end: V) -> float | None:
    """
    Identical to BFS and DFS, except uses a priority queue and weights from the graph.

    Returns the cost of the shortest path between vertices start and end.
    Cost is first in the tuple because heaps are sorted by the first element.

    Runtime: O(|E| + |V| log |V|)
    """
    heap = FibonacciHeap[V]()
    heap.enqueue(start, 0)
    for v in graph:
        if v != start:
            heap.enqueue(v, Graph.INFINITY)
    distances = {}
    while heap:
        # The algorithm guarantees that we now have the shortest distance to u.
        u, cost = heap.dequeue()
        if u == end:
            return cost
        distances[u] = cost
        for v, e in graph[u].items():
            if v not in distances:
                path_cost = cost + e.weight
                if path_cost < heap[v].priority:
                    heap.decrease_key(v, path_cost)
    return None


def dijkstra_shortest_paths(graph: Graph[V], start: V) -> dict[V, float]:
    """
    Dijkstra's algorithm for the single-source shortest paths problem.

    Given a directed, weighted graph G and a source node s, produces the distances from
    s to each other node in the graph. If any nodes in the graph are unreachable from s,
    they will be reported at distance +infinity.

    The code makes up to |E| calls to decrease-key on the heap (worst case, every edge
    from every node yields a shorter path than before) and |V| calls to dequeue-min
    (each node is removed from the heap at most once).

    Runtime: O(|E| + |V| log |V|)
    """
    heap = FibonacciHeap[V]()
    heap.enqueue(start, 0)
    for v in graph:
        if v != start:
            heap.enqueue(v, Graph.INFINITY)
    distances = {}
    while heap:
        # The algorithm guarantees that we now have the shortest distance to u.
        u, cost = heap.dequeue()
        distances[u] = cost
        for v, e in graph[u].items():
            if v not in distances:
                if e.weight < 0:
                    raise RuntimeError("Dijkstra's does not work for negative weights.")
                path_cost = cost + e.weight
                if path_cost < heap[v].priority:
                    heap.decrease_key(v, path_cost)
    return distances


def dijkstra_search_heapq(graph: Graph[V], start: V, end: V) -> float | None:
    """
    Identical to BFS and DFS, except uses a priority queue and weights from the graph.

    Returns the cost of the shortest path between vertices start and end.
    Cost is first in the tuple because heaps are sorted by the first element.

    Runtime: O(|E + V| log |V|)
    """
    heap: list[tuple[float, V]] = [(0, start)]
    visited: set[V] = set()
    while heap:
        cost, u = heapq.heappop(heap)
        if u == end:
            return cost
        if u not in visited:
            visited.add(u)
            for v, e in graph[u].items():
                if v not in visited:
                    heapq.heappush(heap, (cost + e.weight, v))
    return None


def dijkstra_shortest_paths_heapq(graph: Graph[V], start: V) -> dict[V, float]:
    heap: list[tuple[float, V]] = [(0, start)]
    visited: set[V] = set()
    distances = {v: Graph.INFINITY for v in graph}
    distances[start] = 0
    while heap:
        cost, u = heapq.heappop(heap)
        if u not in visited:
            visited.add(u)
            for v, e in graph[u].items():
                if distances[u] + cost < distances[v]:
                    distances[v] = distances[u] + e.weight
                    if v not in visited:
                        heapq.heappush(heap, (cost + e.weight, v))
    return distances



from __future__ import annotations


def kmp_string_match(needle: str, haystack: str) -> int | None:
    """
    The Knuth-Morris-Pratt algorithm finds an occurrence of the specified needle string
    in the haystack string. To do this, we compute a failure table. Next, we iterate
    across the string, keeping track of a candidate start point and length matched so
    far. Whenever a match occurs, we update the length of the match we've made. On a
    failure, we update these values by trying to preserve the maximum proper border
    of the string we were able to manage by that point.

    kmp_string_match("0101", "0011001011") == 5
    """
    # Create the failure table, which for length zero is None.
    fail = [0]
    for i, ch in enumerate(needle):
        # Keep track of the size of the subproblem we're dealing with, which
        # starts off using the first i characters of the string.
        j = i
        while True:
            # If j hits zero, the recursion says that the resulting value is
            # zero since we're looking for the LPB of a single-character string.
            if j == 0:
                fail.append(0)
                break

            # Otherwise, if the character one step after the LPB matches the
            # next character in the sequence, then we can extend the LPB by one
            # character to get an LPB for the whole sequence.
            if needle[fail[j]] == ch:
                fail.append(fail[j] + 1)
                break

            # Finally, if neither of these hold, then we need to reduce the
            # subproblem to the LPB of the LPB.
            j = fail[j]

    # Keep track of the start index and next match position, both of which
    # start at zero since our candidate match is at the beginning and is trying
    # to match the first character.
    index, match = 0, 0
    while index + match < len(haystack):
        # If the current char matches the expected char, bump up the match index.
        if haystack[index + match] == needle[match]:
            match += 1
            if match == len(needle):
                return index

        # Otherwise, we need to look at the fail table to determine what to do next.
        else:
            # If we couldn't match the first character, then just advance the
            # start index.  We need to try again.
            if match == 0:
                index += 1

            # Otherwise, see how much we need to skip forward before we have
            # another feasible match.
            else:
                index += match - fail[match]
                match = fail[match]
    return None



from __future__ import annotations

from collections.abc import Hashable, Iterator
from dataclasses import dataclass
from enum import Enum, auto, unique
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class DisjointSetNode(Generic[T]):
    data: T
    index: int
    rank: int = 0
    size: int = 1

    def __post_init__(self) -> None:
        self.parent = self

    def __hash__(self) -> int:
        return hash(self.data)


@unique
class UnionMode(Enum):
    SIZE = auto()
    RANK = auto()
    INDEX = auto()


@dataclass(init=False, repr=False)
class DisjointSet(Generic[T]):
    """
    Implementation of DisjointSet used in Kruskal's algorithm as a union-find data
    structure. Let # of elements n and a(n) be the inverse Ackermann function.

    Runtime:
        Find: O(a(n))
        Union: O(a(n))

    Space: O(n)
    """

    mode: UnionMode

    def __init__(self, *, mode: UnionMode = UnionMode.INDEX) -> None:
        self.node_data: list[DisjointSetNode[T]] = []
        self.data_to_index: dict[T, int] = {}
        self.mode = mode
        self.num_roots = 0

    def __len__(self) -> int:
        return self.num_roots

    def __contains__(self, item: T) -> bool:
        return item in self.data_to_index

    def __bool__(self) -> bool:
        return bool(self.node_data)

    def __iter__(self) -> Iterator[T]:
        for node in self.node_data:
            yield node.data

    def __getitem__(self, elem: T) -> DisjointSetNode[T]:
        return self.node_data[self.data_to_index[elem]]

    def __repr__(self) -> str:
        return f"DisjointSet({self.itersets()})"

    def make_set(self, elem: T) -> None:
        """
        make x as a set.
        rank is the distance from x to its' parent
        root's rank is 0
        """
        if not isinstance(elem, Hashable):
            raise KeyError(f"Element is not a hashable type: {elem}")
        if elem not in self.data_to_index:
            new_node_index = len(self.node_data)
            node = DisjointSetNode(elem, new_node_index)
            self.node_data.append(node)
            self.data_to_index[elem] = new_node_index
            self.num_roots += 1

    def find_set(self, elem: T) -> T:
        """
        Returns the root parent of x using path splitting, which was proven by
        Patwary, Mostofa Ali et al. to be the most efficient variant.
        https://algocoding.wordpress.com/2015/05/13/simple-union-find-techniques/
        """
        if elem not in self.data_to_index:
            raise KeyError(f"Disjoint set does not contain element: {elem}")
        x = self[elem]
        while x != x.parent:
            x, x.parent = x.parent, x.parent.parent
        return x.data

    def union(self, a: T, b: T) -> None:
        """
        Union two sets. Make the set with bigger rank the parent, so that the
        disjoint set tree will be more flat.
        Returns True if two sets were merges, False otherwise.
        """
        a, b = self.find_set(a), self.find_set(b)
        if a == b:
            return
        x, y = self[a], self[b]

        if self.mode is UnionMode.SIZE:
            if x.size < y.size:
                x, y = y, x
            y.parent = x
            x.size += y.size

        elif self.mode is UnionMode.RANK:
            if x.rank < y.rank:
                x, y = y, x
            y.parent = x
            if x.rank == y.rank:
                x.rank += 1

        elif self.mode is UnionMode.INDEX:
            if self.data_to_index[a] < self.data_to_index[b]:
                x.parent = y.parent
            else:
                y.parent = x.parent

        self.num_roots -= 1

    def is_connected(self, x: T, y: T) -> bool:
        """
        :param x: first element
        :param y: second element
        :return: True if x and y belong to the same tree
            (i.e. they have the same root), False otherwise.
        """
        return self.find_set(x) == self.find_set(y)

    def itersets(self) -> list[set[T]]:
        """
        Yields sets of connected components.
        The roots dict shares the set references with entries in the same set.
        """
        roots: dict[T, set[T]] = {}
        seen_nodes: dict[DisjointSetNode[T], T] = {}
        for node in self.node_data:
            curr = node

            unseen_nodes = set()
            cached = False
            while curr != curr.parent:
                if curr in seen_nodes:
                    root = seen_nodes[curr]
                    roots[root].add(node.data)
                    cached = True
                    break
                unseen_nodes.add(curr)
                curr = curr.parent

            root = curr.data
            for new_node in unseen_nodes:
                seen_nodes[new_node] = root

            if not cached:
                if root not in roots:
                    roots[root] = set()
                roots[root].add(node.data)
        return list(roots.values())

    def naive_itersets(self) -> list[set[T]]:
        """
        Yields sets of connected components.
        The roots dict shares the set references with entries in the same set.
        """
        roots: dict[T, set[T]] = {}
        for node in self.node_data:
            curr = node
            while curr != curr.parent:
                curr = curr.parent

            root = curr.data
            if root not in roots:
                roots[root] = set()
            roots[root].add(node.data)
        return list(roots.values())



from __future__ import annotations

from typing import Any, Iterable, Mapping, cast

from cs.structures import Edge, Graph, V


def bipartite_matching(
    graph: Graph[V], left: Iterable[V] | Mapping[V, int], right: Iterable[V]
) -> tuple[Graph[V], list[Edge[V]]]:
    source = cast(Any, "Source")
    sink = cast(Any, "Sink")
    graph.add_node(source)
    graph.add_node(sink)
    if isinstance(left, Mapping):
        for node, count in left.items():
            graph.add_edge(source, node, weight=count)
    else:
        for node in left:
            graph.add_edge(source, node, weight=1)
    for node in right:
        graph.add_edge(node, sink)

    max_flow_graph = ford_max_flow_network(graph, source, sink)
    matching_edges = [
        edge
        for edge in max_flow_graph.edges
        if edge["flow"] >= 1 and edge.start in left and edge.end in right
    ]
    return max_flow_graph, matching_edges


def ford_max_flow(graph: Graph[V], source: V, sink: V) -> float:
    result = ford_max_flow_network(graph, source, sink)
    return sum(edge["flow"] for edge in result[source].values())


def ford_max_flow_network(graph: Graph[V], source: V, sink: V) -> Graph[V]:
    """
    Given a graph and a pair of nodes s and t, produces a maximum s-t flow in that
    graph. Flow that already exists in the input network will be used as a guess of
    the maximum flow.

    The algorithm works by maintaining a candidate max flow, constructing the residual
    graph for that max flow, and then searching for an augmenting s-t path in the graph.
    If one is found, then the maximum possible flow is pushed along that path and the
    process repeats. Otherwise, no augmenting paths exist, and the flow must be maximum.

    Each time an augmenting path is found, the s-t flow increases by at least one, so
    the algorithm is guaranteed to terminate after at most F searches, where F is the
    value of the max-flow. Since each search to find an augmenting path from s to t
    takes O(m + n), the overall runtime is O((m + n)F).

    Runtime: O((m + n)F)
    """
    flow_network = Graph.from_graph(
        graph, edge_fn=lambda e: Edge(**e, capacity=e.weight, flow=0)
    )
    if source == sink:
        return flow_network

    # Pushing flow across an edge in the original graph is equivalent to retracting flow
    # across an edge in the residual graph. The flow across an edge is represented
    # implicitly by the amount of capacity remaining on its residual edge.
    residual_graph = Graph.from_graph(
        flow_network,
        edge_fn=lambda e: Edge(
            e.start, e.end, capacity=e["capacity"] - e["flow"], is_residual=False
        ),
    )
    # Add all of the reverse edges with capacity equal to the reverse flow.
    for edge in flow_network.edges:
        residual_graph.add_edge(
            edge.end, edge.start, capacity=edge["flow"], is_residual=True
        )

    while True:
        # Find an augmenting s-t path in the residual graph.
        path = find_path(residual_graph, source, sink)
        if path is None:
            break

        # Augments the flow along an augmenting path by the minimum remaining capacity
        # along some edge. Determine the smallest capacity of this path, then push this
        # much flow across all the nodes.
        min_capacity = min(edge["capacity"] for edge in path)
        for edge in path:
            add_flow(residual_graph, edge, min_capacity)

    # We now have a max flow because no augmenting paths are left.  Take the data from
    # our residual graph and use it to fill in the flow in the resulting flow network.
    for u in residual_graph:
        for v, edge in residual_graph[u].items():
            if edge["is_residual"]:
                flow_network[v][u]["flow"] = edge["capacity"]

    return flow_network


def add_flow(graph: Graph[V], edge: Edge[V], amount: int) -> None:
    """
    Adds or subtracts the indicated number of flow units across this edge. If the amount
    of flow added or removed exceeds the capacity of the edge, throws a RuntimeError.
    """
    if amount < 0:
        add_flow(graph, graph[edge.end][edge.start], -amount)
        return

    capacity = edge["capacity"]
    if amount > capacity:
        raise RuntimeError(
            f"Cannot push {amount} units of flow across edge of capacity {capacity}."
        )
    edge["capacity"] -= amount
    graph[edge.end][edge.start]["capacity"] += amount


def find_path(
    residual_graph: Graph[V], start: V, end: V, visited: set[V] | None = None
) -> list[Edge[V]] | None:
    """
    Recursively explores a residual graph, starting at the node indicated by start and
    searching for a particular destination node. If a path is found, it is returned.
    """
    if visited is None:
        visited = set()

    if start in visited:
        return None

    visited.add(start)
    if start == end:
        return []

    for v, edge in residual_graph[start].items():
        if edge["capacity"] == 0:
            continue

        result = find_path(residual_graph, v, end, visited)
        if result is not None:
            result.append(edge)
            return result

    return None



from typing import TypeVar

from cs.util import Comparable

T = TypeVar("T", bound=Comparable)


def selection_sort(array: list[T]) -> list[T]:
    """
    Selection sort algorithm implementation.

    Runtime: O(n^2)
    """
    for i in range(len(array) - 1):
        least = i
        for k in range(i + 1, len(array)):
            if array[k] < array[least]:
                least = k
        if least != i:
            array[least], array[i] = array[i], array[least]
    return array
